////////////////////////////////////////////////////////
//
//	VGAPaint 386 - Animation I/O routines.
//
//	Loading and saving VAN, ANI, and the like.
//
//	Codec for Microsoft Video 1 AVI files derived from that of
//	Xanim270 by Mark Podlipec.  Here is the copyright notice:

	/*
	 * xa_avi.c
	 *
	 * Copyright (C) 1993,1994,1995,1996 by Mark Podlipec. 
	 * All rights reserved.
	 *
	 * This software may be freely copied, modified and redistributed without
	 * fee for non-commerical purposes provided that this copyright notice is
	 * preserved intact on all copies and modified copies.
	 * 
	 * There is no warranty or other guarantee of fitness of this software.
	 * It is provided solely "as is". The author(s) disclaim(s) all
	 * responsibility and liability with respect to this software's usage
	 * or its effect upon hardware or computer systems.
	*/
 
#define	VGAPAINT_ANIMIO_C

#include <string.h>

#include <a/bits.h>

#include "io.h"
#include "animio.h"
#include "picture.h"
#include "video.h"

#include "average.h"

///////////////////////////////////////////////////////////

struct VPTemporal1CInfo {
	ulong	bfsiz;		// size of bitfield
	ulong	tlsiz;		// size of raw tile block
	ulong	min_size;	// absolute minimum frame size
	ulong	max_size;	// absolute maximum frame size
	int	quality;	// starting quality setting
	boolean	adaptive;	// use adaptive system?
	boolean	drift;		// let quality drift?
};

#pragma off (check_stack)

static uchar *TDbridiff;

int	TileDiff(uchar *tile1,uchar *tile2,ulong bpr) {
	return	  (ulong)TDbridiff[(tile1[0*bpr+0]<<8)+tile2[0*bpr+0]]
		+ (ulong)TDbridiff[(tile1[0*bpr+1]<<8)+tile2[0*bpr+1]]
		+ (ulong)TDbridiff[(tile1[0*bpr+2]<<8)+tile2[0*bpr+2]]
		+ (ulong)TDbridiff[(tile1[0*bpr+3]<<8)+tile2[0*bpr+3]]
		+ (ulong)TDbridiff[(tile1[1*bpr+0]<<8)+tile2[1*bpr+0]]
		+ (ulong)TDbridiff[(tile1[1*bpr+1]<<8)+tile2[1*bpr+1]]
		+ (ulong)TDbridiff[(tile1[1*bpr+2]<<8)+tile2[1*bpr+2]]
		+ (ulong)TDbridiff[(tile1[1*bpr+3]<<8)+tile2[1*bpr+3]]
		+ (ulong)TDbridiff[(tile1[2*bpr+0]<<8)+tile2[2*bpr+0]]
		+ (ulong)TDbridiff[(tile1[2*bpr+1]<<8)+tile2[2*bpr+1]]
		+ (ulong)TDbridiff[(tile1[2*bpr+2]<<8)+tile2[2*bpr+2]]
		+ (ulong)TDbridiff[(tile1[2*bpr+3]<<8)+tile2[2*bpr+3]]
		+ (ulong)TDbridiff[(tile1[3*bpr+0]<<8)+tile2[3*bpr+0]]
		+ (ulong)TDbridiff[(tile1[3*bpr+1]<<8)+tile2[3*bpr+1]]
		+ (ulong)TDbridiff[(tile1[3*bpr+2]<<8)+tile2[3*bpr+2]]
		+ (ulong)TDbridiff[(tile1[3*bpr+3]<<8)+tile2[3*bpr+3]];
}

#define	ACT_PUT(v,bits) do { 				\
		if (accbits<(bits)) {			\
			*bitptr++ = (accum<<accbits)|((v)>>((bits)-accbits));	\
			accum = (v) & ((1L<<((bits)-accbits))-1);	\
			accbits = 8 + accbits - (bits);		\
		} else {			\
			accum <<= (bits); 	\
			accum |= (v);		\
			accbits -= (bits);	\
		}	\
	} while(0);

static int AnimCompTemporal1(struct VPTemporal1CInfo *results,struct BitMap *tfbm,struct BitMap *lfbm,struct BitMap *srcbm,
	int quality,uchar *bitmem,uchar *tilemem) {

	uchar	accum,accbits=8;
	int	xt = srcbm->width >> 2;
	int	yt = srcbm->height >> 2;

	uchar	*srcc = srcbm->vm;		// current frame source data
	uchar	*srcp = lfbm->vm;		// last frame source data
	uchar	*dst = tfbm->vm;		// current frame destination data
	ulong	smod = srcbm->bpr*4 - xt*4;
	ulong	dmod = tfbm->bpr*4 - xt*4;
	ulong	bpr = srcbm->bpr;
	uchar	*bitptr = bitmem;
	uchar	*tilptr = tilemem;

	int	x,y;

	y = yt;
	do {
		x = xt;
		do {

		#define	CD_CNLAST	(0)
		#define CD_UPLAST	(1)
		#define	CD_DNLAST	(2)
		#define	CD_LFLAST	(3)
		#define	CD_RTLAST	(4)
		#define	CD_LFTHIS	(5)
		#define	CD_UPTHIS	(6)

			long	bd,bdmax;
			uchar	code;

			bdmax = TileDiff(srcc,srcp,bpr);
			code = CD_CNLAST;

			if (y<yt) {
				if ((bd = TileDiff(srcc,srcp-4*bpr,bpr))<bdmax) { bdmax = bd; code = CD_UPLAST; }
				if ((bd = TileDiff(srcc,srcc-4*bpr,bpr))<bdmax) { bdmax = bd; code = CD_UPTHIS; }
			}
			if (x!=xt) {
				if ((bd = TileDiff(srcc,srcp-4,bpr))<bdmax) { bdmax = bd; code = CD_LFLAST; }
				if ((bd = TileDiff(srcc,srcc-4,bpr))<bdmax) { bdmax = bd; code = CD_LFTHIS; }
			}
			if (y>1)
				if ((bd = TileDiff(srcc,srcp+4*bpr,bpr))<bdmax) { bdmax = bd; code = CD_DNLAST; }
			if (x>1)
				if ((bd = TileDiff(srcc,srcp+4,bpr))<bdmax) { bdmax = bd; code = CD_RTLAST; }

			if (bdmax<quality) {
				switch(code) {
				case CD_CNLAST:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(srcp + 0*bpr);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(srcp + 1*bpr);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(srcp + 2*bpr);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(srcp + 3*bpr);
					ACT_PUT(1,2);
					break;
				case CD_UPLAST:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(srcp - 4*bpr);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(srcp - 3*bpr);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(srcp - 2*bpr);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(srcp - 1*bpr);
					ACT_PUT(12,4);
					break;
				case CD_DNLAST:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(srcp + 4*bpr);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(srcp + 5*bpr);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(srcp + 6*bpr);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(srcp + 7*bpr);
					ACT_PUT(13,4);
					break;
				case CD_LFLAST:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(srcp + 0*bpr - 4);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(srcp + 1*bpr - 4);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(srcp + 2*bpr - 4);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(srcp + 3*bpr - 4);
					ACT_PUT(14,4);
					break;
				case CD_RTLAST:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(srcp + 0*bpr + 4);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(srcp + 1*bpr + 4);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(srcp + 2*bpr + 4);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(srcp + 3*bpr + 4);
					ACT_PUT(15,4);
					break;
				case CD_LFTHIS:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(dst + 0*bpr - 4);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(dst + 1*bpr - 4);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(dst + 2*bpr - 4);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(dst + 3*bpr - 4);
					ACT_PUT(4,3);
					break;
				case CD_UPTHIS:
					*(ulong *)(dst + 0*bpr) = *(ulong *)(dst - 4*bpr);
					*(ulong *)(dst + 1*bpr) = *(ulong *)(dst - 3*bpr);
					*(ulong *)(dst + 2*bpr) = *(ulong *)(dst - 2*bpr);
					*(ulong *)(dst + 3*bpr) = *(ulong *)(dst - 1*bpr);
					ACT_PUT(5,3);
					break;
				}
			} else {
				*(ulong *)(tilptr+ 0) = *(ulong *)(dst + 0*bpr) = *(ulong *)(srcc + 0*bpr);
				*(ulong *)(tilptr+ 4) = *(ulong *)(dst + 1*bpr) = *(ulong *)(srcc + 1*bpr);
				*(ulong *)(tilptr+ 8) = *(ulong *)(dst + 2*bpr) = *(ulong *)(srcc + 2*bpr);
				*(ulong *)(tilptr+12) = *(ulong *)(dst + 3*bpr) = *(ulong *)(srcc + 3*bpr);
				tilptr += 16;

				ACT_PUT(0,2);
			}

			srcc += 4;
			srcp += 4;
			dst += 4;
		} while(--x);
		srcc	+= smod;
		srcp	+= smod;
		dst	+= dmod;
	} while(--y);

	if (accbits < 8)
		*bitptr++ = accum << accbits;

	results->bfsiz = bitptr - bitmem;
	results->tlsiz = tilptr - tilemem;

	mouse_hide();
	mouse_show();

	return results->bfsiz + results->tlsiz + 4;
}

///////////////////////////////////////////////////////////
//
//	enum PictIOErrors AnimWriteVP(char *fname, struct Picture *pict,boolean use_temporal1,boolean use_delta2);
//
//	Writes an animation out in VP386 format.

enum PictureIOErrors AnimWriteVP(char *fname,struct Picture *pict,char *audio_fname,boolean use_temporal1,boolean use_delta2) {
	FILE *f,*faudio=NULL;
	struct VPAnimHeader vpah;
	struct VPFrameHeader vpfh;
	struct VPNewPalette vppl;
	struct VPAnimAudioHeader vpaah;
	struct VPTemporal1CInfo vpci;
	struct FramePalette *fpal;
	ulong	last_audiopos=0;
	ulong	new_audiopos;
	uchar	*audio_copybuf=NULL;
	int	audio_sequence = 0;
	ulong	xtiles,ytiles;
	int i;
	ulong	lastsize=0;

	///////

	struct BitMap lfbm;		// last frame
	struct BitMap tfbm;		// this frame
	uchar	*lfvm = NULL;		// last frame memory
	uchar	*bitmem;
	uchar	*tilemem;

	///////

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;

	// allocate a buffer

	if (use_temporal1) {
		ulong	sz;

		if ((pict->bm->width & 3) || (pict->bm->height & 3)) {
			Message("Anim width and height must be multiples of 4");
			return PIOERR_UNKNOWN;
		}
		xtiles = pict->bm->width>>2;
		ytiles = pict->bm->height>>2;

		sz = vd->rassize(pict->bm->width,pict->bm->height);
		if (!(lfvm = malloc(262144+sz*3+((xtiles*ytiles*4+7)>>3))))
			return PIOERR_MEMORY;

		TDbridiff	= lfvm;
		tilemem = lfvm + 262144 + sz*2;
		bitmem	= lfvm + 262144 + sz*3;

		vid_initBitMap(&lfbm,lfvm+262144   ,0,0,pict->bm->width,pict->bm->height,pict->bm->bpr,pict->bm->pal);
		vid_initBitMap(&tfbm,lfvm+262144+sz,0,0,pict->bm->width,pict->bm->height,pict->bm->bpr,pict->bm->pal);
		vd->set(&lfbm,4);
		vd->set(&tfbm,4);

		{
			uchar r1,g1,b1;
			uchar r2,g2,b2;
			int i,j;

			r1=g1=b1=0;
			for(j=0;j<252;j++) {
				r2 = g2 = b2 = 0;
				for(i=0;i<252;i++) {
					ulong d;
					long d2;

					d2 = r2-r1;	d = d2*d2;
					d2 = g2-g1;	d += d2*d2;
					d2 = b2-b1;	d += d2*d2;
					TDbridiff[((j+4)<<8) + (i+4)] = d;
					if (++b2>=6) { b2=0; if (++g2>=7) { g2=0; ++r2; }}
				}
				if (++b1>=6) { b1=0; if (++g1>=7) { g1=0; ++r1; }}
			}
		}

		if (vpci.adaptive = !!RequestBoolean("VP386 Temporal 1 Options","Yes",NULL,"No","Use Adaptive compression?")) {
			vpci.drift = RequestBoolean("VP386 Temporal 1 Options","Yes",NULL,"No","Let quality drift?");
			vpci.min_size = RequestSlider("Minimum frame size",0,65536,512);
			vpci.max_size = RequestSlider("Maximum frame size",-vpci.min_size,65536,max(8192,12288-vpci.min_size));
			vpci.quality = RequestSlider("Initial quality",0,1024,10);
		}
	}

	if (!(f=VPfopen(fname,"wb"))) {
		free(lfvm);
		return PIOERR_OPEN;
	}

	// fill AnimHeader, write out

	memset(&vpah,0,sizeof vpah);
	vpah.magic		= VPAH_MAGIC;
	vpah.len		= sizeof(struct VPAnimHeader);
	vpah.frames		= pict->total_frames;
	vpah.playback_rate	= pict->playback_speed;
	vpah.width		= pict->bm->width;
	vpah.height		= pict->bm->height;
	vpah.flags		= NULL;
	memcpy(vpah.palette,pict->palette,768);
	memcpy(vpah.palbits,pict->pal_bits,32);

	if (audio_fname) {
		if (!(faudio = VPfopen(audio_fname,"rb"))) {
			free(lfvm);
			fclose(f);
			return PIOERR_OPEN;
		}
		if (-1L==(i = VPfilesize(faudio))) goto read_error;

		vpah.audio_len = i;
		vpah.audio_rate = 11025;
		vpah.audio_channels = 1;

		// allocate copybuf memory

		if (!(audio_copybuf = malloc(16384))) goto memory_error;

	}

	if (1!=fwrite(&vpah,sizeof vpah,1,f)) goto write_error;

	// for each frame, dump out frame header & data

	if (!PictureFlush(pict)) goto memory_error;
	if (!PictureGotoFrame(pict,pict->first,1)) goto memory_error;

	fpal = pict->first->frame_pal;

	for(i=0;i<pict->total_frames;i++) {

		// write out audio header if we need to

		if (audio_fname) {
			new_audiopos = scale32u((i+1)*vpah.audio_rate,vpah.playback_rate,65536L)+65536;
			if (new_audiopos > last_audiopos) {
				int bytes;

				last_audiopos += 16384;

				if ((bytes=fread(audio_copybuf,1,16384,faudio))<0) goto read_error;

				if (bytes) {
					vpaah.magic	= VPAAH_MAGIC;
					vpaah.len	= sizeof(struct VPAnimAudioHeader)+bytes;
					vpaah.sequence	= audio_sequence++;
					vpaah.channel	= 0;
					vpaah.pan	= 0;
					vpaah.volume	= 255;
					vpaah.sample_len= bytes;

					if (1!=fwrite(&vpaah,sizeof vpaah,1,f)) goto write_error;
					if (1!=fwrite(audio_copybuf,bytes,1,f)) goto write_error;
				} else audio_fname = NULL;
			}
		}

		if (pict->current->frame_pal != fpal) {
			if (pict->current->frame_pal) {
				vppl.magic = 'LPPV';
				vppl.len = sizeof(struct VPNewPalette);
				memcpy(vppl.palette,pict->current->frame_pal->palette,768);
				if (1!=fwrite(&vppl,sizeof vppl,1,f)) goto write_error;
			} else
				if (1!=fwrite("VPPL\0\0\0",8,1,f)) goto write_error;

			fpal = pict->current->frame_pal;
		}

		vpfh.magic		= VPFH_MAGIC;
		vpfh.len		= sizeof(struct VPFrameHeader);

		if (use_temporal1) {
			ulong t;
			int q;
			boolean was_too_big;

			q= vpci.quality;
			was_too_big = FALSE;

			t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);

			if (t > vpci.max_size) {
				do {
					q += vpci.drift ? 1 : 16;
					t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);
					SetTitle("Frame %d: quality %3d size %7ld last %7ld",i+1,q,t,lastsize);
				} while(t>vpci.max_size);

				// now tighten it a bit

				do {
					if (q>1) --q; else break;
					t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);
					SetTitle("Frame %d: quality %3d size %7ld last %7ld",i+1,q,t,lastsize);
				} while(t<vpci.max_size);

				++q;
				t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);
			} else if (t<vpci.min_size) {
				do {
					if (q > 1) --q;
					else break;
					t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);
					SetTitle("Frame %d: quality %3d size %7ld last %7ld",i+1,q,t,lastsize);
				} while(t<vpci.min_size);
				if (t>vpci.max_size) {
					++q;
					t = AnimCompTemporal1(&vpci,&tfbm,&lfbm,pict->bm,q,bitmem,tilemem);
				}
			}
			SetTitle("Frame %d: quality %3d size %7ld last %7ld",i+1,q,t,lastsize);

			lastsize = t;

			if (vpci.drift) vpci.quality = q;

			vd->bitblt(&tfbm,0,0,&lfbm,0,0,lfbm.width,lfbm.height);
			vd->bitblt(&tfbm,0,0,&vidbm,0,11,lfbm.width,lfbm.height);

			vpfh.flags		= VPFH_COMP_TEMPORAL1;
			vpfh.framesiz		= vpci.bfsiz + vpci.tlsiz + 4;
			if (1!=fwrite(&vpfh,sizeof vpfh,1,f)) goto write_error;

			if (1!=fwrite(&vpci.bfsiz,4,1,f)) goto write_error;
			if (1!=fwrite(bitmem,vpci.bfsiz,1,f)) goto write_error;
			if (vpci.tlsiz)
				if (1!=fwrite(tilemem,vpci.tlsiz,1,f)) goto write_error;
		} else if (use_delta2) {
			uchar *dm1,*dm2;
			int x,y;
			uchar c,d;
			ushort encode_table[256];
			uchar bitlen_table[256];
			ushort bl_encode_table[16];
			uchar bl_bitlen_table[16];
			ulong size1,size2;

			// double-delta the picture

			dm1 = pict->bm->vm;

			y = pict->bm->height;
			do {
				dm2=dm1;
				dm1+=pict->bm->bpr;
				x = pict->bm->width-1;
				c = *dm2++;
				do {
					d = *dm2;
					*dm2++ = d-c;
					c=d;
				} while(--x);
			} while(--y);

			dm1 = pict->bm->vm;

			x = pict->bm->width;
			do {
				dm2=dm1++;
				y = pict->bm->height-1;
				c = *dm2;	dm2 += pict->bm->bpr;
				do {
					d = *dm2;
					*dm2 = d-c;
					dm2 += pict->bm->bpr;
					c=d;
				} while(--y);
			} while(--x);

			// call the Huffman libraries and make a huffman table for the deltas

			if ((size1 = HuffmanMakeEncodingTable(pict->bm->vm,pict->bm->size,encode_table,bitlen_table,256))<0)
				goto memory_error;

			if ((size2 = HuffmanMakeEncodingTable(bitlen_table,256,bl_encode_table,bl_bitlen_table,16))<0)
				goto memory_error;

			vpfh.flags		= VPFH_COMP_DELTA2HUFFMAN;
			vpfh.framesiz		= size1+size2+8;
			if (1!=fwrite(&vpfh,sizeof vpfh,1,f)) goto write_error;

			HuffmanWriteNibbleTable(f,bl_bitlen_table,16);
			HuffmanWriteBlock(f,bitlen_table,256,bl_encode_table,bl_bitlen_table);
			HuffmanWriteBlock(f,pict->bm->vm,pict->bm->size,encode_table,bitlen_table);
		} else {
			vpfh.flags		= pict->current->compressed ? VPFH_COMP_BYTERUN : VPFH_COMP_NONE;
			vpfh.framesiz		= pict->current->frame_siz;
			if (1!=fwrite(&vpfh,sizeof vpfh,1,f)) goto write_error;
			if (1!=fwrite(pict->current->frame_mem,pict->current->frame_siz,1,f)) goto write_error;
		}

		// the picture should NOT be marked dirty at this point

		PictureNextFrame(pict);

		if (-1 != kbd_check()) break;
	}

	if (f) fclose(f);
	if (faudio) fclose(faudio);
	if (lfvm) free(lfvm);
	if (audio_copybuf) free(audio_copybuf);
	return PIOERR_NONE;
write_error:
	if (f) fclose(f);
	if (faudio) fclose(faudio);
	if (lfvm) free(lfvm);
	if (audio_copybuf) free(audio_copybuf);
	return PIOERR_WRITE;
memory_error:
	if (f) fclose(f);
	if (faudio) fclose(faudio);
	if (lfvm) free(lfvm);
	if (audio_copybuf) free(audio_copybuf);
	return PIOERR_MEMORY;
read_error:
	if (f) fclose(f);
	if (faudio) fclose(faudio);
	if (lfvm) free(lfvm);
	if (audio_copybuf) free(audio_copybuf);
	return PIOERR_READ;
}


//	enum PictureIOErrors AnimReadVP(struct Picture *pict,char *fname,int insert_mode);
//
//	Loads in a VP386 format Anim file and inserts it into/replaces the current picture.

enum PictureIOErrors AnimReadVP(struct Picture *pict,char *fname,int insert_mode) {
	FILE *f;
	struct VPAnimHeader vpah;
	struct VPFrameHeader vpfh;
	struct FramePalette *fpal = NULL;
	int i;

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;
	if (!(f=VPfopen(fname,"rb"))) return PIOERR_OPEN;

	if (1!=fread(&vpah,sizeof vpah.magic+sizeof vpah.len,1,f))		goto read_error;
	if (vpah.magic != VPAH_MAGIC)			goto bad_format;	// not a VP386 animation

	// seems okay... read in the rest

	if (1!=fread(&vpah.frames,vpah.len-(sizeof vpah.magic+sizeof vpah.len),1,f)) goto read_error;

//	if (vpah.len < sizeof(struct VPAnimHeader))	goto bad_format;	// VPAnimHeader too short

	if (vpah.len > sizeof(struct VPAnimHeader))			// skip extra VPAnimHeader data
		if (fseek(f,vpah.len-sizeof(struct VPAnimHeader),SEEK_SET)) goto read_error;

	// check width and height - if not right, request user for Anim resize/abort

	if (vpah.width != pict->bm->width  ||  vpah.height != pict->bm->height) {
		if (RequestBoolean("VP386 Anim Warning","Resize & Load",NULL,"Abort Load",
				"Current Anim is %ldx%ld - File is %ldx%ld",pict->bm->width,pict->bm->height,
					vpah.width,vpah.height) != 2)
			goto user_abort;

		// resize & destroy anim

		if (!PictureReinit(pict,vpah.width,vpah.height)) goto no_memory;
	}

	// initialize picture fields from file's animation fields

	memcpy(pict->palette,vpah.palette,768);
	memcpy(pict->pal_bits,vpah.palbits,32);
	pict->playback_speed = vpah.playback_rate;

	// start loading the animation in, frame by frame

	for(i=vpah.frames;i;i--) {
		for(;;) {
			if (1!=fread(&vpfh.magic,(sizeof vpfh.magic)+(sizeof vpfh.len),1,f)) goto read_error;
				// error reading VPFrameHeader

//		if (vpfh.magic != VPFH_MAGIC) goto bad_format;		// this frame sucks

			if (vpfh.magic == VPPL_MAGIC) {	// new frame palette
				if (vpfh.len) {
					if (!(fpal = CreateFramePalette())) goto no_memory;
					AttachPaletteToPicture(pict_cur,fpal);
					if (1!=fread(fpal->palette,min(768,vpfh.len-8),1,f)) goto read_error;

					// take up any excess

					if (vpfh.len > sizeof(struct VPNewPalette))
						if (fseek(f,vpfh.len-sizeof(struct VPNewPalette),SEEK_CUR)) goto read_error;
				} else fpal = NULL;
			} else if (vpfh.magic != VPFH_MAGIC) {
				if (fseek(f,vpfh.len-sizeof vpfh.magic-sizeof vpfh.len,SEEK_CUR)) goto read_error;
			} else break;
		}
		if (1!=fread(&vpfh.flags,sizeof vpfh-sizeof vpfh.magic-sizeof vpfh.len,1,f)) goto read_error;	// error reading VPFrameHeader
		if (vpfh.len < sizeof(struct VPFrameHeader)) goto bad_format;	// header too short
		else if (vpfh.len > sizeof(struct VPFrameHeader))
			if (fseek(f,vpfh.len-sizeof(struct VPFrameHeader),SEEK_CUR)) goto read_error;

		// header checks out - insert a frame and decompress to it

		if (i<vpah.frames) if (!InsertAfterFrame(pict)) goto no_memory;		// couldn't insert frame

		// decompress to the frame buffer

		switch(vpfh.flags & VPFH_COMPMASK) {
		case VPFH_COMP_NONE:
			if (!AllocateFrameBuffer(pict->current,vpfh.framesiz,FALSE)) goto no_memory;
			if (1!=fread(pict->current->frame_mem,vpfh.framesiz,1,f)) goto read_error;
			break;
		case VPFH_COMP_BYTERUN:
			if (!AllocateFrameBuffer(pict->current,vpfh.framesiz,TRUE)) goto no_memory;
			if (1!=fread(pict->current->frame_mem,vpfh.framesiz,1,f)) goto read_error;
			break;
		default:
			goto bad_format;		// #&*@$^*&@# bad compression
		}

		// attach frame palette

		if (fpal) AttachPaletteToFrame(pict,pict->current,fpal);

		PictureInvalidate(pict);
	}

	// the last frame is still in memory... advance to the next frame to flush it

	PictureNextFrame(pict);

	// success!  close and exit

	fclose(f);
	return PIOERR_NONE;

read_error:	fclose(f);	return PIOERR_READ;
bad_format:	fclose(f);	return PIOERR_BADFORMAT;
user_abort:	fclose(f);	return PIOERR_ABORT_USER;
no_memory:	fclose(f);	return PIOERR_MEMORY;
}

//	enum PictureIOErrors AnimReadPCX(struct Picture *pict,char *fname,int insert_mode);
//
//	Loads in a series of PCX files and inserts it into/replaces the current picture.

enum PictureIOErrors AnimReadPCX(struct Picture *pict,char *fname,int insert_mode) {
	FILE *f;
	char *s,*t;
	int err;

	s = fname;
	while(*s) ++s;
	--s;
	while(s>fname)
		if (*s--=='.') break;

	if (s[1]!='.')
		return PIOERR_UNKNOWN;

	while(f=VPfopen(fname,"rb")) {
		fclose(f);

		// attach a palette to the frame

		DetachFramePalette(pict,pict->current);
		AttachPaletteToFrame(pict,pict->current,NULL);

		if (PIOERR_NONE!=(err=PCXRead(fname,1)))
			return err;
		InsertAfterFrame(pict);

		// try to advance filename

		t=s;
		while(t>=fname) {
			if (isdigit(*t)) {
				if (*t=='9')
					*t--='0';
				else {
					(*t)++;
					break;
				}
			} else {
				*t = '0';
				break;
			}
		}
	}
	PictureNextFrame(pict);

	return PIOERR_NONE;
}

//	enum PictureIOErrors AnimWritePCX(struct Picture *pict,char *fname);
//
//	Saves a sequence of PCX pictures.

enum PictureIOErrors AnimWritePCX(struct Picture *pict,char *fname) {
	int err;
	int fr;
	char *s,*t;
	struct StatusBar *sb;

	PictureGotoFrame(pict,pict->first,1);

	s = fname;
	while(*s) ++s;
	--s;
	while(s>fname)
		if (*s--=='.') break;

	if (s[1]!='.')
		return PIOERR_UNKNOWN;

	sb = CreateStatusBar(pict->total_frames,"Saving %ld PCX files...",pict->total_frames);

	for(fr=pict->total_frames;fr;fr--) {
		if (PIOERR_NONE != (err = PCXWrite(fname,picbm,pict_cur->cur_pal))) {
			DestroyStatusBar(sb);
			return err;
		}

		ModifyStatusBar(sb,pict->total_frames-fr);

		// advance filename

		t=s;
		while(t>=fname) {
			if (isdigit(*t)) {
				if (*t=='9')
					*t--='0';
				else {
					(*t)++;
					break;
				}
			} else {
				*t = '0';
				break;
			}
		}
		if (t<fname) {
			Message("Out of filename characters.");
			break;
		}
		if (!PictureNextFrame(pict)) {
			DestroyStatusBar(sb);
			return PIOERR_MEMORY;
		}
	}

	DestroyStatusBar(sb);
	return PIOERR_NONE;
}

////////////////////////////////////////////////////////////

struct RIFFheader {
	ulong	id;
	ulong	size;
};

struct AVIheader {
	ulong	dwMicroSecPerFrame;
	ulong	dwMaxBytesPerSec;
	ulong	dwReserved1;
	ulong	dwFlags;
	ulong	dwTotalFrames;
	ulong	dwInitialFrames;
	ulong	dwStreams;
	ulong	dwSuggestedBufferSize;
	ulong	dwWidth;
	ulong	dwHeight;
	ulong	dwScale;
	ulong	dwRate;
	ulong	dwStart;
	ulong	dwLength;
};


struct AVIStreamHeader {
	ulong	fccType;
	ulong	fccHandler;
	ulong	dwFlags;
	ulong	dwReserved1;
	ulong	dwInitialFrames;
	ulong	dwScale;
	ulong	dwRate;
	ulong	dwStart;
	ulong	dwLength;
	ulong	dwSuggestedBufferSize;
	ulong	dwQuality;
	ulong	dwSampleSize;
};

#define RIFF_RIFF	'FFIR'
#define RIFF_LIST	'TSIL'
#define	RIFF_AVI	' IVA'
#define	RIFF_rec	' cer'
#define	RIFF_hdrl	'lrdh'
#define RIFF_strl	'lrts'
#define RIFF_strh	'hrts'
#define RIFF_strf	'frts'
#define	RIFF_vids	'sdiv'
#define	RIFF_auds	'sdua'
#define	RIFF_avih	'hiva'
#define	RIFF_movi	'ivom'
#define	RIFF_01wb	'bw10'
#define	RIFF_00dc	'cd00'
#define	RIFF_00db	'bd00'

// <rgb>ptr must point to the first character of the last line

boolean	AVIdecompress_DIB16(uchar *rptr,uchar *gptr,uchar *bptr,uchar *src,ulong width,ulong height,ulong bpr) {
	rptr += (height-1)*bpr;
	gptr += (height-1)*bpr;
	bptr += (height-1)*bpr;

	while(height--) {
		ushort d;
		int x;

		for(x=0;x<width;x++) {
			d = *(ushort *)src; src+=2;

			bptr[x] = expand5to8( d     &0x1f);
			gptr[x] = expand5to8((d>> 5)&0x1f);
			rptr[x] = expand5to8((d>>10)&0x1f);
		}
		bptr -= bpr;
		gptr -= bpr;
		rptr -= bpr;

		// pad to longword
		src += (4-2*width)&3;
	}

	return TRUE;
}

boolean	AVIdecompress_DIB24(uchar *rptr,uchar *gptr,uchar *bptr,uchar *src,ulong width,ulong height,ulong bpr) {
	rptr += (height-1)*bpr;
	gptr += (height-1)*bpr;
	bptr += (height-1)*bpr;

	while(height--) {
		int x;

		for(x=0;x<width;x++) {
			bptr[x] = *src++;
			gptr[x] = *src++;
			rptr[x] = *src++;
		}
		bptr -= bpr;
		gptr -= bpr;
		rptr -= bpr;

		// pad to longword

		src += (4-3*width)&3;
	}

	return TRUE;
}

#define	AVI_DROP_24(f1,fv,o,v) \
	if ((f1) & (fv)) { rptr[o] = rB##v; gptr[o] = gB##v; bptr[o] = bB##v; } \
	else		 { rptr[o] = rA##v; gptr[o] = gA##v; bptr[o] = bA##v; }

boolean	AVIdecompress_CRAM16(uchar *rptr,uchar *gptr,uchar *bptr,uchar *src,ulong width,ulong height,ulong bpr) {
	int y,x;

	rptr += (height-4)*bpr;
	gptr += (height-4)*bpr;
	bptr += (height-4)*bpr;

	width = (width+3) & ~3;
	y = ((height+3)>>2)-1;
	x = 0;

	while(y>=0) {
		uchar c0,c1;

		c0 = *src++; c1 = *src++;

		if (c1 >= 0x84 && c1 <= 0x87) {
			int skip = ((ulong)(c1-0x84)<<8) + (ulong)c0;

			x += skip;
			rptr += 4*skip;
			gptr += 4*skip;
			bptr += 4*skip;
		} else {
			uchar rA0,gA0,bA0;

			if (c1<0x80) {	// 2 or 8 color encoding
				uchar rB0,gB0,bB0;
				ushort a,b;

				b = *(ushort *)(src + 0);
				a = *(ushort *)(src + 2); src += 4;

				rA0 = expand5to8((a>>10)&0x1f);	gA0 = expand5to8((a>>5)&0x1f);	bA0 = expand5to8(a&0x1f);
				rB0 = expand5to8((b>>10)&0x1f);	gB0 = expand5to8((b>>5)&0x1f);	bB0 = expand5to8(b&0x1f);

				if (b & 0x8000) {	// 8 color encoding
					uchar rA1,rB1,gA1,gB1,bA1,bB1;

					b = *(ushort *)(src + 0);
					a = *(ushort *)(src + 2); src += 4;

					rA1 = expand5to8((a>>10)&0x1f);	gA1 = expand5to8((a>>5)&0x1f);	bA1 = expand5to8(a&0x1f);
					rB1 = expand5to8((b>>10)&0x1f);	gB1 = expand5to8((b>>5)&0x1f);	bB1 = expand5to8(b&0x1f);

					AVI_DROP_24(c0,0x01,3*bpr+0,0);
					AVI_DROP_24(c0,0x02,3*bpr+1,0);
					AVI_DROP_24(c0,0x04,3*bpr+2,1);
					AVI_DROP_24(c0,0x08,3*bpr+3,1);
					AVI_DROP_24(c0,0x10,2*bpr+0,0);
					AVI_DROP_24(c0,0x20,2*bpr+1,0);
					AVI_DROP_24(c0,0x40,2*bpr+2,1);
					AVI_DROP_24(c0,0x80,2*bpr+3,1);

					b = *(ushort *)(src + 0);
					a = *(ushort *)(src + 2);
					rA0 = expand5to8((a>>10)&0x1f);	gA0 = expand5to8((a>>5)&0x1f);	bA0 = expand5to8(a&0x1f);
					rB0 = expand5to8((b>>10)&0x1f);	gB0 = expand5to8((b>>5)&0x1f);	bB0 = expand5to8(b&0x1f);

					b = *(ushort *)(src + 4);
					a = *(ushort *)(src + 6); src += 8;
					rA1 = expand5to8((a>>10)&0x1f);	gA1 = expand5to8((a>>5)&0x1f);	bA1 = expand5to8(a&0x1f);
					rB1 = expand5to8((b>>10)&0x1f);	gB1 = expand5to8((b>>5)&0x1f);	bB1 = expand5to8(b&0x1f);

					AVI_DROP_24(c1,0x01,1*bpr+0,0);
					AVI_DROP_24(c1,0x02,1*bpr+1,0);
					AVI_DROP_24(c1,0x04,1*bpr+2,1);
					AVI_DROP_24(c1,0x08,1*bpr+3,1);
					AVI_DROP_24(c1,0x10,0*bpr+0,0);
					AVI_DROP_24(c1,0x20,0*bpr+1,0);
					AVI_DROP_24(c1,0x40,0*bpr+2,1);
					AVI_DROP_24(c1,0x80,0*bpr+3,1);
				} else {		// 2 color encoding
					AVI_DROP_24(c0,0x01,3*bpr+0,0);
					AVI_DROP_24(c0,0x02,3*bpr+1,0);
					AVI_DROP_24(c0,0x04,3*bpr+2,0);
					AVI_DROP_24(c0,0x08,3*bpr+3,0);
					AVI_DROP_24(c0,0x10,2*bpr+0,0);
					AVI_DROP_24(c0,0x20,2*bpr+1,0);
					AVI_DROP_24(c0,0x40,2*bpr+2,0);
					AVI_DROP_24(c0,0x80,2*bpr+3,0);
					AVI_DROP_24(c1,0x01,1*bpr+0,0);
					AVI_DROP_24(c1,0x02,1*bpr+1,0);
					AVI_DROP_24(c1,0x04,1*bpr+2,0);
					AVI_DROP_24(c1,0x08,1*bpr+3,0);
					AVI_DROP_24(c1,0x10,0*bpr+0,0);
					AVI_DROP_24(c1,0x20,0*bpr+1,0);
					AVI_DROP_24(c1,0x40,0*bpr+2,0);
					AVI_DROP_24(c1,0x80,0*bpr+3,0);
				}
			} else {		// 1 color encoding
				ushort a;

				a = ((c1<<8)+c0);
				rA0 = expand5to8((a>>10)&0x1f);	gA0 = expand5to8((a>>5)&0x1f);	bA0 = expand5to8(a&0x1f);

				*(ulong *)(rptr + 0*bpr) =
				*(ulong *)(rptr + 1*bpr) =
				*(ulong *)(rptr + 2*bpr) =
				*(ulong *)(rptr + 3*bpr) = expand8to32(rA0);

				*(ulong *)(gptr + 0*bpr) =
				*(ulong *)(gptr + 1*bpr) =
				*(ulong *)(gptr + 2*bpr) =
				*(ulong *)(gptr + 3*bpr) = expand8to32(gA0);

				*(ulong *)(bptr + 0*bpr) =
				*(ulong *)(bptr + 1*bpr) =
				*(ulong *)(bptr + 2*bpr) =
				*(ulong *)(bptr + 3*bpr) = expand8to32(bA0);
			}

			rptr += 4;
			gptr += 4;
			bptr += 4;
			x++;
		}

		while(x>=(width>>2)) {
			x-=(width>>2);
			y--;
			rptr -= width + bpr*4;
			gptr -= width + bpr*4;
			bptr -= width + bpr*4;
		}
	}

	return TRUE;
}

enum PictureIOErrors AnimReadAVI(struct Picture *pict,char *fname,int insert_mode) {
	FILE *f,*fs;
	struct RIFFheader rh;
	struct AVIheader avih;
	struct AVIStreamHeader avish;
	struct AVIStreamHeader v_avish;
	struct BMPInfoHeader bmph;
	ulong	tag,len,tc,l;
	uchar	*copymem=NULL;
	ulong	copymem_siz=0;
	uchar *sepmem = NULL;
	boolean	last_was_video = FALSE;

	if (!(f=VPfopen(fname,"rb")))
		return PIOERR_OPEN;

	if (!(sepmem = malloc(picbm->size*3))) {
		fclose(f);
		return PIOERR_MEMORY;
	}

	while(1==fread(&rh,sizeof rh,1,f) && -1==kbd_check()) {
		switch(rh.id) {
		case RIFF_RIFF:
			if (1!=fread(&tag,sizeof tag,1,f)) {
				Message("couldn't read RIFF tag");
				goto unknown;
			}
			rh.size-=4;
			switch(tag) {
			case RIFF_AVI:
				break;
			default:
				if (fseek(f,(rh.size+1)&~1,SEEK_CUR)) {
					Message("couldn't seek");
					goto unknown;
				}
			}
			break;
		case RIFF_LIST:
			if (1!=fread(&tag,sizeof tag,1,f)) {
				Message("couldn't read RIFF tag");
				goto unknown;
			}
			rh.size-=4;
			switch(tag) {
			case RIFF_rec:
				break;
			case RIFF_hdrl:
				break;
			case RIFF_movi:
				break;
			case RIFF_strl:
				break;
			default:
				if (fseek(f,(rh.size+1)&~1,SEEK_CUR)) {
					Message("couldn't seek");
					goto unknown;
				}
			}
			break;
		case RIFF_avih:
			if (1!=fread(&avih,sizeof avih,1,f)) {
				Message("couldn't read");
				goto unknown;
			}
			if (fseek(f,(-rh.size)&1,SEEK_CUR)) {
				Message("couldn't seek");
				goto unknown;
			}

//			Message("avi size: %dx%d\n",avih.dwWidth,avih.dwHeight);

			if (avih.dwWidth!=pict->bm->width || avih.dwHeight!=pict->bm->height) {
				Message("Picture is %ldx%ld; AVI is %ldx%ld",pict->bm->width,pict->bm->height,avih.dwWidth,avih.dwHeight);
				goto unknown;
			}

			break;
		case RIFF_00dc:
		case RIFF_00db:
			l = (rh.size+1)&~1;

			if (copymem_siz < l)
				if (copymem = realloc(copymem,l))
					copymem_siz = l;
				else
					goto mem;

			if (1!=fread(copymem,l,1,f)) {
				Message("couldn't read");
				goto unknown;
			}

			SetTitle("%ld",pict->current_num);

			DetachFramePalette(pict,pict->current);
			AttachPaletteToFrame(pict,pict->current,NULL);

			{
				int y,x;
				uchar *src;
				uchar *rptr,*gptr,*bptr;

				src = copymem;

				rptr = sepmem + picbm->size*0;
				gptr = sepmem + picbm->size*1;
				bptr = sepmem + picbm->size*2;

				switch(v_avish.fccHandler) {
				case ' BID':
					switch(bmph.biBitCount) {
					case 24:	AVIdecompress_DIB24(rptr,gptr,bptr,src,avih.dwWidth,avih.dwHeight,picbm->width);
							break;
					case 16:	AVIdecompress_DIB16(rptr,gptr,bptr,src,avih.dwWidth,avih.dwHeight,picbm->width);
							break;
					}
					break;
				case 'cvsm':
					AVIdecompress_CRAM16(rptr,gptr,bptr,src,avih.dwWidth,avih.dwHeight,picbm->bpr);
					break;
				}
			}

			ObtainPalette2(sepmem,sepmem+picbm->size,sepmem+picbm->size*2,FALSE,reduce_maxcol);
			Separate_Remap2(sepmem,sepmem+picbm->size,sepmem+picbm->size*2,FALSE);
			InsertAfterFrame(pict);

			break;
		case RIFF_strh:
			if (1!=fread(&avish,sizeof avish,1,f)) {
				Message("couldn't read");
				goto unknown;
			}

			if (last_was_video = (avish.fccType == RIFF_vids)) {
				memcpy(&v_avish,&avish,sizeof avish);

				// we can handle 16-bit and 24-bit "DIB "s, along with 16-bit "CRAM"s

				if (avish.fccHandler != ' BID' && avish.fccHandler != 'cvsm') {
					Message("AVI video is neither DIB nor CRAM format");
					goto unknown;
				}
			}

			break;
		case RIFF_strf:
			if (last_was_video) {
				if (1!=fread(&bmph,sizeof bmph,1,f)) {
					Message("couldn't read");
					goto unknown;
				}

				Message("bits: %d",bmph.biBitCount);

				// if it's a DIB (uncompressed), it can be either 16-bit or 24-bit
				// if it's a CRAM (Microsoft Video 1), it can only be 16-bit at this time

				switch(v_avish.fccHandler) {
				case ' BID':
					if (bmph.biBitCount != 16 && bmph.biBitCount != 24) {
						Message("AVI DIB files must be either 16 or 24 bit");
						goto unknown;
					}
					break;
				case 'cvsm':
					if (bmph.biBitCount != 16) {
						Message("AVI CRAM files must be 16-bit");
						goto unknown;
					}
					break;
				}
			} else {
				if (fseek(f,(rh.size+1)&~1,SEEK_CUR)) {
					Message("couldn't seek");
					goto unknown;
				}
			}
			break;
		default:
			if (fseek(f,(rh.size+1)&~1,SEEK_CUR)) {
				Message("couldn't seek");
				goto unknown;
			}
		}
	}

	free(copymem);
	free(sepmem);
	fclose(f);
	return PIOERR_NONE;
unknown:
	free(copymem);
	free(sepmem);
	fclose(f);
	return PIOERR_UNKNOWN;
mem:
	free(sepmem);
	fclose(f);
	return PIOERR_MEMORY;
}

////////////////////////////////////////////////////////////

// "ACONfram"
//	(count){"icon"<length>{data}}
//	"anih"<36>{36 bytes}
//	"rate"{length of rate block} {data}
//	"seq" {length of sequence block} {data}

struct CURCursorDir {
	short	cdReserved;
	short	cdType;
	short	cdCount;
};

struct CURCursorDirEntry {
	char	bWidth;
	char	bHeight;
	char	bColorCount;
	char	bReserved;
	short	wXHotspot;
	short	wYHotspot;
	long	lBytesInRes;
	long	dwImageOffset;
};

struct ANIHeader {
	ulong	cbSizeOf;
	ulong	cFrames;
	ulong	cSteps;
	ulong	cx,cy;
	ulong	cBitCount,cPlanes;
	ulong	JifRate;
	ulong	flags;
};

#define ANIF_ICON (0x0001L) // Windows format icon/cursor animation

enum PictureIOErrors AnimWriteANI(struct Picture *pict,char *fname) {
	FILE	*f;
	ulong	ary[256];
	int	i,j,k;
	struct ANIHeader anih;
	struct CURCursorDir ccd;
	struct CURCursorDirEntry ccde;
	struct BMPInfoHeader bmpih;
	char *fpalette;

	////

	memset(&bmpih,0,sizeof bmpih);

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;
	if (!(f=VPfopen(fname,"wb"))) return PIOERR_OPEN;

	if (!PictureFlush(pict)) goto abort_memory;
	if (!PictureGotoFrame(pict,pict->first,1)) goto abort_memory;

	ary[0] = 'FFIR';
	ary[1] = 104+(778+512+240*4)*pict->total_frames;
	ary[2] = 'NOCA';
	ary[3] = 'TSIL';
	ary[4] = 28;
	ary[5] = 'OFNI';
	ary[6] = 'MANI';
	ary[7] = 4;
	ary[8] = 'a';
	ary[9] = 'TRAI';
	ary[10]= 4;
	ary[11]= 'a';
	if (1!=fwrite(ary,48,1,f)) goto abort_write;

	anih.cbSizeOf	= sizeof(struct ANIHeader);
	anih.cFrames	= pict->total_frames;
	anih.cSteps	= pict->total_frames;
	anih.cx		= 0;
	anih.cy		= 0;
	anih.cBitCount	= 4;
	anih.cPlanes	= 1;
	anih.JifRate	= 6;
	anih.flags	= ANIF_ICON;

	ary[0] = 'hina';
	ary[1] = 36;
	if (1!=fwrite(ary,8,1,f)) goto abort_write;
	if (1!=fwrite(&anih,sizeof anih,1,f)) goto abort_write;

	ary[0] = 'etar';
	ary[1] = sizeof(ulong)*pict->total_frames;
	if (1!=fwrite(ary,8,1,f)) goto abort_write;
	for(i=0;i<pict->total_frames;i++) {
		ary[0] = 6;
		if (1!=fwrite(ary,4,1,f)) goto abort_write;
	}

	ary[0] = 'TSIL';
	ary[1] = 4+(774+512+240*4)*pict->total_frames;
	ary[2] = 'marf';
	if (1!=fwrite(ary,12,1,f)) goto abort_write;

	for(i=0;i<pict->total_frames;i++) {
		fpalette = pict->current->frame_pal ? pict->current->frame_pal->palette : pict->palette;

		ary[0] = 'noci';
		ary[1] = 766+512+240*4;
		if (1!=fwrite(ary,8,1,f)) goto abort_write;

		ccd.cdReserved		= 0;
		ccd.cdType		= 2;
		ccd.cdCount		= 1;
		ccde.bWidth		= 32;
		ccde.bHeight		= 32;
		ccde.bColorCount	= 0;
		ccde.bReserved		= 0;
		ccde.wXHotspot		= 0;
		ccde.wYHotspot		= 0;
		ccde.lBytesInRes	= 744-512+1024+240*4;
		ccde.dwImageOffset	= 0x16;
		bmpih.biSize		= sizeof bmpih;
		bmpih.biWidth		= 32;
		bmpih.biHeight		= 64;
		bmpih.biPlanes		= 1;
		bmpih.biBitCount	= 8;	//4;
		bmpih.biSizeImage	= 0x280 + 512;
		bmpih.biClrUsed		= 256;
		bmpih.biClrImportant	= 256;

		if (1!=fwrite(&ccd,sizeof ccd,1,f)) goto abort_write;		//  6 bytes (6)
		if (1!=fwrite(&ccde,sizeof ccde,1,f)) goto abort_write;		// 16 bytes (22)
		if (1!=fwrite(&bmpih,sizeof bmpih,1,f)) goto abort_write;	// 40 bytes (62)

		for(j=0;j<256;j++) {
			((uchar *)(ary+j))[0] = fpalette[3*j+2];
			((uchar *)(ary+j))[1] = fpalette[3*j+1];
			((uchar *)(ary+j))[2] = fpalette[3*j+0];
		}
//		memset(ary+8,0,sizeof(ulong)*8);
		if (1!=fwrite(ary,sizeof(ulong)*256,1,f)) goto abort_write;	// 64 bytes (126)

		{
			uchar *t = pict->bm->vm+31*32;

			for(k=0;k<32;k++) {
#if 0
				for(j=0;j<16;j++) {
					putc((t[0]&15)|(t[1]<<4),f);
					t+=2;
				}
#else
				fwrite(t,32,1,f);
				t+=32;
#endif
				t-=64;
			}
		}

		memset(ary,0,sizeof(ulong)*16);
		for(j=0;j<2;j++) if (1!=fwrite(ary,sizeof(ulong)*16,1,f)) goto abort_write;

		if (!PictureNextFrame(pict)) goto abort_memory;
	}

	if (fclose(f)) return PIOERR_WRITE;
	return PIOERR_NONAME;

abort_write:
	fclose(f);
	return PIOERR_WRITE;
abort_memory:
	fclose(f);
	return PIOERR_MEMORY;
}

////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
//
//	enum PictIOErrors AnimWriteFLC(struct Picture *pict,char *fname);
//
//	Writes an animation out in uncompressed FLC format.

enum PictureIOErrors AnimWriteFLC(struct Picture *pict,char *fname) {
	FILE *f = NULL;
	int i;
	int j;
	struct FLCHeader flch;
	struct FLCBlock flcb;
	struct FLCSubBlock flcsb;

	///////

	if (!fname) return PIOERR_NONAME;
	if (!*fname) return PIOERR_NONAME;

	if (!(f=VPfopen(fname,"wb"))) return PIOERR_OPEN;

	// fill AnimHeader, write out

	memset(&flch,0,sizeof flch);
	flch.total_size		= 128 + 6 + 771 + (pict->bm->size + 22) * pict->total_frames;
	flch.id			= FLC_ID;
	flch.frame_count	= pict->total_frames;
	flch.width		= pict->bm->width;
	flch.height		= pict->bm->height;
	flch.bitsperpixel	= 8;
	flch.speed		= scale32u(pict->playback_speed,125L,8192);
	flch.creator		= '83PV';
	flch.modifier		= '83PV';
	flch.x_aspect		= 1;
	flch.y_aspect		= 1;
	flch.offset_first	= 128;
	flch.offset_second	= 128 + 6 + 771 + pict->bm->size + 22;

	if (1!=fwrite(&flch,sizeof flch,1,f)) goto write_error;

	// for each frame, dump out frame header & data

	if (!PictureFlush(pict)) goto memory_error;
	if (!PictureGotoFrame(pict,pict->first,1)) goto memory_error;

	for(i=0;i<pict->total_frames;i++) {
		memset(&flcb,0,sizeof flcb);
		flcb.size	= pict->bm->size + 22;	if (!i) flcb.size += 771 + 6;
		flcb.id		= FLCB_ID;
		flcb.subblocks	= i ? 1 : 2;
		if (1!=fwrite(&flcb,sizeof flcb,1,f)) goto write_error;

		if (!i) {
			flcsb.size	= 771 + 6;
			flcsb.type	= FLI_COLOR256;
			if (1!=fwrite(&flcsb,sizeof flcsb,1,f)) goto write_error;
			fputc(1,f);
			fputc(0,f);
			fputc(0,f);
			for(j=0;j<256;j++) {
				fputc(pict->palette[j*3+2],f);
				fputc(pict->palette[j*3+1],f);
				fputc(pict->palette[j*3+0],f);
			}
		}

		flcsb.size	= 6 + pict->bm->size;
		flcsb.type	= FLI_COPY;
		if (1!=fwrite(&flcsb,sizeof flcsb,1,f)) goto write_error;
		if (1!=fwrite(pict->bm->vm,pict->bm->size,1,f)) goto write_error;

		PictureNextFrame(pict);

		if (-1 != kbd_check()) break;
	}

	if (f) fclose(f);
	return PIOERR_NONE;
write_error:
	if (f) fclose(f);
	return PIOERR_WRITE;
memory_error:
	if (f) fclose(f);
	return PIOERR_MEMORY;
}

