#ifndef VGAPAINT_CAPTURE_H
#define VGAPAINT_CAPTURE_H

#include <a/types.h>

#ifdef	VGAPAINT_CAPTURE_C
#define	INIT(x)	=x
#define	EXTERN
#else
#define	INIT(x)
#define	EXTERN	extern
#endif

enum	VGACaptureType {
	VCT_SIMPLE	=0,
	VCT_RF		=1,
	VCT_RFX		=2,
	VCT_RFC		=3,
	VCT_ADVANCED	=4,
	VCT_VESA	=5
};

struct	VGACaptDynaChain {
	struct VGACaptDynaChain *prev,*next;
};

struct	VGACaptureMode {
	ulong	flags;		// various flags
	uchar	joyMask;	// joystick port trigger mask
	uchar	lastJoyRead;	// last joyport buttons
	uchar	type;		// capture mode
	uchar	pad;
	ulong	max_static;	// amount of static memory allocated to capture
	ulong	max_dynamic;	// amount of dynamic memory (DPMI) allocated to capture
	ulong	cur_static;	// amount of static memory currently in use
	ulong	cur_dynamic;	// amount of dynamic memory currently in use
	ulong	frames;		// number of frames captured
	ulong	framesbad;	// number of frames *missed* due to a lack of mem
	ulong	framesiz;	// bytes per frame - not used in Advanced mode
	uchar	*static_mem;	// pointer to static memory
	uchar	*static_ptr;	// current pointer
	struct VGACaptDynaChain	*dyna_chain;	// dynamic memory chain
	ulong	last_chk;	// rapid-fire: "last checksum"
	ulong	*last_ptr;	// rapid-fire: last frame
	boolean	rapidfire;	// rapid-fire, meaning skip triggers

	// dynamic memory is not yet supported
};

EXTERN	struct VGACaptureMode	vgacm;

struct VGACapture {
	struct VGACapture *next;
	uchar miscout;		// miscellaneous output
	uchar palmask;		// palette mask register
	uchar seqreg;		// sequencer index register
	uchar crtcreg;		// crtc index register
	uchar graphreg;		// graph index register
	uchar seq[5];		// sequencer registers
	uchar crtc[0x19];	// CRTC registers
	uchar graph[0x09];	// Graphics controller registers
	uchar attr[0x15];	// Attribute controller registers
	uchar palette[768];	// color palette
	uchar plane[4][65536];	// video memory -- very large!
};

// CaptureInterpret() fills in this info

enum	VGAPixelType	{ CDPT_16P=0,CDPT_4CP=1,CDPT_256BPP=2 };

enum	VGACaptureFlags {
	CDF_ALPHA	= 0x00000001L
};

struct VGACaptureInfo {
	ulong		flags;		// flags
	short		pwidth;		// pixel width
	short		pheight;	// pixel height
	short		plinecmp;	// pixel line compare position
	uchar		pix_type;	// pixel type
	uchar		pix_mult;	// number of pixels per CRTC address
	uchar		pix_height;	// height of a pixel in scan lines
};

EXTERN struct VGACapture *capture;

void	CaptureDisplay(void);
void	Capture_Install(void);
void	Capture_Deinstall(void);
void	Capture_FreeMemory(void);
void	Capture_Print_Info(void);
void	CaptureMutatePalette(uchar *);
void	CaptureDisplayRegs(void);

#undef EXTERN
#undef INIT

#endif

