/****** WARNING: THIS MODULE CONTAINS INTERRUPT ROUTINES *********/
/****** THESE ROUTINES MUST USE SS != DS/ES *********/

#define VGAPAINT_CCYCLE_C
#include <mem.h>
#include <dos.h>

#include <a/types.h>
#include <a/joy.h>
#include <a/bits.h>

#include "ccycle.h"
#include "capture.h"
#include "io.h"
#include "picture.h"

static struct IHandler CCycle_Int={
	{ {NULL},(IntFuncPtr)&CCycler,NULL },
	{ NULL },
	0,
	0,
};

static int CCycle_Inhibit_Cnt=0;

void CCycle_Suspend(void) {
#if 0
	if (++CCycle_Inhibit_Cnt==1)
		vid_loadRGB6(0,piccol,256);
#else
	++CCycle_Inhibit_Cnt;
#endif
}

void CCycle_Resume(void) {
	if (CCycle_Inhibit_Cnt)
		--CCycle_Inhibit_Cnt;
}

void CCycle_Install(void) {
	if (!ccycle_instd) {
		ccycle_instd=TRUE;
		cyclehalt=FALSE;
		CCycle_Reset();
//		prev_int_1c=_dos_getvect(0x1c);
//		_dos_setvect(0x1c,&CCycle_Interrupt);

		AddInterruptHandler(INT_TIMER,&CCycle_Int);
	};
}

void CCycle_Remove(void) {
	if (ccycle_instd) {
		ccycle_instd=FALSE;
//		_dos_setvect(0x1c,prev_int_1c);

		RemInterruptHandler(INT_TIMER,&CCycle_Int);

		vid_loadRGB6(0,piccol,256);
	};
}

void CCycle_Toggle(void) {
	if (ccycle_instd) CCycle_Remove();
		else      CCycle_Install();
}

void CCycle_Reset(void) {
	volatile uchar *s=cyccol;
	uchar *t=piccol;
	long c=256;

	cyclehalt=TRUE;

	if (opt.gamma_adjust)
		do {
			*s++=monitor_gamma[*t++];
			*s++=monitor_gamma[*t+++256];
			*s++=monitor_gamma[*t+++512];
		} while(--c);
	else {
		c*=3;
		do {
			*s++ = *t++ >> 2;
		} while(--c);
	}

	cyclehalt=FALSE;
}

uchar *CCycle_Unravel(short ccrange) {
	struct CycleRange *cr=&cyclerange[ccrange];
	struct CycleStrip *cs;
	short scnt;
	short slen=0;
	uchar index;
	schar direction;
	uchar *smem,*s2;

	cs=&cr->strip[0];
	scnt=cr->strips;
	while(scnt--) {
		slen+=abs((short)cs->end-(short)cs->start)+1;
		++cs;
	};
	
	if (!(smem=malloc(slen+2))) return NULL;
	s2=smem+2;
	cs=&cr->strip[0];
	scnt=cr->strips;
	while(scnt--) {
		direction=cs->start<cs->end ? 1 : -1;
		index=cs->start;
		for(;;) {
			*s2++=index;
			if (index==cs->end) break;
			index+=direction;
		};
		++cs;
	};

	*(short*)smem=slen;
	return smem;
}

//	void CCycle_Make_Translate_Table(uchar *table,boolean forward);
//
//	Make a translation table that rotates all colors in each range
//	forward or backward one step.

void	CCycle_Make_Translate_Table(uchar *table,boolean forward) {
	struct CycleRange *cr = pict_cur->cyclerange;
	struct CycleStrip *cs;
	int	i,strip,s,t;
	uchar	c;

	for(i=0;i<256;i++) table[i] = i;

	for(i=0;i<8;i++) {
		cs = cr->strip;

		if (cr->strips) {
			if (forward) {
				t = table[cr->strip[cr->strips-1].end];
				for(strip=0;strip<cr->strips;strip++) {
					s = table[cs->end];

					if (cs->start < cs->end)
						memmove(table+cs->start+1,table+cs->start,cs->end-cs->start);
					else if (cs->start > cs->end)
						memmove(table+cs->end,table+cs->end+1,cs->start-cs->end);

					table[cs->start] = t;
					t = s;
					++cs;
				}
			} else {
				t = table[cr->strip[cr->strips-1].start];
				for(strip=0;strip<cr->strips;strip++) {
					s = table[cs->start];

					if (cs->end < cs->start)
						memmove(table+cs->end+1,table+cs->end,cs->start-cs->end);
					else if (cs->end > cs->start)
						memmove(table+cs->start,table+cs->start+1,cs->end-cs->start);

					table[cs->end] = t;
					t = s;
					++cs;
				}
			}
		}
		++cr;
	}
}

//static char prtsc_routine[]={
//	0x00,				/* flag */
//	0x2e,0x80,0x0e,0x00,0x00,0x01,	/* or cs:byte ptr [0],1 */
//	0xcd,0x12,			/* int 12h */
//	0xcf				/* iret */
//};

void	DoCapture(void);

void PrtSc_Install(void) {
	if (prtsc_instd) return;

	if (!(prtsc_callback = AllocCallback((VPpassup_fptr)&DoCapture,TRUE))) return;

//	/* create a far jump in real space */
//
//	memcpy(realmem,prtsc_routine,sizeof prtsc_routine);
//
	regs.x.eax=0x0200;	/* read real-mode interrupt vector */
	regs.h.bl=0x05;		/* Print Screen vector */
	int386(0x31,&regs,&regs);
	prev_int_05_segment=regs.w.dx;
	prev_int_05_offset =regs.w.cx;
	regs.x.eax=0x0201;	/* set real-mode interrupt vector */
	regs.h.bl=0x05;		/* Print Screen vector */
	regs.w.cx=prtsc_callback;
	regs.w.dx=0;
	int386(0x31,&regs,&regs);
//
//	prev_int_12=_dos_getvect(0x12);
//	_dos_setvect(0x12,&PrtSc_Interrupt);

	prtsc_instd=TRUE;
}

void PrtSc_Remove(void) {
	if (!prtsc_instd) return;
	regs.x.eax=0x0201;	/* set real-mode interrupt vector */
	regs.h.bl=0x05;		/* Print Screen vector */
	regs.w.cx=prev_int_05_segment;
	regs.w.dx=prev_int_05_offset;
	int386(0x31,&regs,&regs);

//	_dos_setvect(0x12,prev_int_12);
	FreeCallback(prtsc_callback);
	prtsc_callback=0;

	prtsc_instd=FALSE;
}

#pragma off (check_stack)

//#pragma aux fix_regs="push ds","pop es","cld";

/* remember that since SS!=ES/DS, using t as auto would create far ptr */

void CCycler(void) {
	static char s[3];
	static char t[3];
	short cnt;
	short scnt;
	struct CycleRange *cr=cyclerange;
	struct CycleStrip *cs;
	boolean changed=FALSE;
	int	first_changed,last_changed;

	if (CCycle_Inhibit_Cnt) return;

	first_changed = 255;
	last_changed=0;

	cnt=8;
	do {
		if (cr->speed) {
			cr->timer-=Int08TimerPeriod;
			while (cr->timer<0) {
//				cr->timer+=cr->speed;
				cr->timer=cr->speed;
				cs=&cr->strip;
				scnt=cr->strips;
				if (scnt) {
					qmoveb(t,cyccol+cr->strip[scnt-1].end*3,3);
					do {
						qmoveb(s,cyccol+cs->end*3,3);
						if (cs->start < cs->end) {
							if (cs->start < first_changed) first_changed = cs->start;
							if (cs->end > last_changed) last_changed = cs->end;
							memmove(cyccol+cs->start*3+3,cyccol+cs->start*3,(cs->end-cs->start)*3);
						} else if (cs->start > cs->end) {
							if (cs->end < first_changed) first_changed = cs->end;
							if (cs->start > last_changed) last_changed = cs->start;
							memmove(cyccol+cs->end*3,cyccol+cs->end*3+3,(cs->start-cs->end)*3);
						}
						qmoveb(cyccol+cs->start*3,t,3);
						qmoveb(t,s,3);
						cs++;
					} while(--scnt);
				};
				changed=TRUE;
			};
		};
		++cr;
	} while(--cnt);

/* LoadRGB6 re-enables interrupts, so it cannot be used */
	if (changed) {
		outportb(0x3c8,first_changed);
       		outportbs(0x3c9,(uchar *)cyccol + 3*first_changed,3*(last_changed+1-first_changed));
	};
}

//#pragma aux CCycle_Interrupt loadds far;
//void __far __loadds __interrupt CCycle_Interrupt(void) {
//	fix_regs();
//
//	if (!cyclehalt) CCycler();
//	if (!(inportb(0x200) & 0x10)) {
//		outportb(0x61,(inportb(0x61) & 0xfe) ^ 0x02);
//		if (opt.VESACapt) VESA_Capture(); else PrtSc_Interrupt2();
//	};
//
//	_chain_intr(prev_int_1c);
// 
//	/* chain_intr ABORTS - do not do anything here */
//}


///////////////////////////////// capture routines ////////////////////////////////////

//	void *ptr = Capture_Allocate_Memory(ulong size);
//
//	Allocates memory from either static capture memory or DPMI (dynamic capture
//	memory).  CALLED FROM CAPTURE INTERRUPT (Int8).
//
//	Longword alignment is assured.
//
//	Returns NULL if no memory is available.  Note that it is the CALLER's
//	responsibility to change the image captured/missed counts.

void *Capture_Allocate_Memory(ulong size) {
	void *mem;

	// round size up to lword

	size = (size+3) & ~3;

	if ((vgacm.max_static - vgacm.cur_static) < size) {
		// attempt to allocate dynamic memory
		// oops... dynamic memory not implemented yet!

		return NULL;		// no more mem
	};

	// allocate static memory

	mem = vgacm.static_ptr;	vgacm.static_ptr += size;
	vgacm.cur_static += size;

	return mem;
}

//	void Capture_Deallocate_Memory(ulong size);
//
//	Unallocates a chunk of memory recently allocated from static or dynamic
//	memory.  CALLED FROM CAPTURE INTERRUPT (Int8).
//
//	Longword alignment is assured.

void	Capture_Deallocate_Memory(ulong size) {

	// round size up to lword

	size = (size+3) & ~3;

	// deallocate static memory

	vgacm.static_ptr -= size;
	vgacm.cur_static -= size;
}


void Capture_Advanced(void) {
	short x;
	uchar *pp;
	short ioa;
	uchar	dacState;
	uchar	dacColor;
	uchar	dacSubColor=0;

	_disable();		// make SURE of this!

	// Normally, interrupts are disabled automatically after an interrupt occurs by the CPU.
	// However, this doesn't occur if a PUSHF/CALL is used to call the old int handler -
	// this is often used to split timer 0 in two, a fast handler (such as a VBL) and the
	// slow 18Hz timer used by ordinary BIOS routines.  The fast handler re-enables the interrupt,
	// does its work, and calls the slow interrupt handlers if it's time.  If the interrupt
	// resignals, it does its work again, sees a set re-entrancy flag, and exits immediately without
	// re-calling the slow interrupt.  The problem is that programs can modify the VGA registers
	// in interrupt, which we're trying to snag.
	//
	// Disabling interrupts manually fixes this problem, and stops the garbage problem in DOOM(tm)
	// and Heretic(tm).  The IRETD used to end the handler will restore the interrupt flag under
	// normal circumstances (I'm not sure about this - I know POPF doesn't do it)

	// grab all the registers we can first
	// make sure we grab the index regs, too

	if ((capture->miscout=inportb(0x3cc)) & 1)
		ioa=0x3d0;
	else
		ioa=0x3b0;
	capture->palmask=inportb(0x3c6);

//	goto skip_all;

	capture->seqreg  = inportb(0x3c4);
	capture->crtcreg = inportb(ioa+4);
	capture->graphreg= inportb(0x3ce);
	pp=capture->seq;   for(x=0;x< 5;x++) { outportb(0x3c4,x); *pp++=inportb(0x3c5); };
	pp=capture->crtc;  for(x=0;x<25;x++) { outportb(ioa+4,x); *pp++=inportb(ioa+5); };
	pp=capture->graph; for(x=0;x< 9;x++) { outportb(0x3ce,x); *pp++=inportb(0x3cf); };
	pp=capture->attr;  for(x=0;x<21;x++) { inportb(ioa+0x0A); outportb(0x3c0,x); *pp++=inportb(0x3c1); };
	inportb(ioa+0x0A);
	outportb(0x3c0,32);

//	goto skip_all;

	// read the RAMDAC's current counter.  Determine the RAMDAC state and use
	// it to determine whether we are not in the middle of an operation, or
	// whether one byte or two has been transacted...
	//
	// We run into a bad situation here when a write is in progress: we must
	// end the current write operation, but we can't write anything or we overwrite
	// existing colors.  The solution is to switch to read mode in the middle
	// of the operation.  I'm not sure if this will work on IBM VGAs - it's definitely
	// illegal according to the specs.

	outportb(0x3c6,0xFF);		// set palette mask to 0xFF

	dacColor = inportb(0x3c8);
	if (dacState = (inportb(0x3c7) & 3)) {
		// DAC in read op
		do { inportb(0x3c9); ++dacSubColor; } while(inportb(0x3c8)==dacColor);
		--dacColor;	// for some reason...
	} else {
		// DAC in write op
		do { inportb(0x3c9); ++dacSubColor; } while(inportb(0x3c8)==dacColor);
	}
	dacSubColor = 3-dacSubColor;

	pp=capture->palette;
	outportb(0x3c7,0);
	inportbs(0x3c9,pp,768);
//	x=768; do { *pp++=(85L*inportb(0x3c9))/21; } while(--x);

	// restore the RAMDAC counter

	if (dacState) {
		// DAC in read op
		outportb(0x3c7,dacColor);
		while(dacSubColor--) inportb(0x3c9);
	} else {
		uchar *pp2=capture->palette + dacColor*3;

		// DAC in write op

//		outportb(0x3c8,dacColor);	// restore the color we trashed earlier
//		outportbs(0x3c9,pp2, 3);

		outportb(0x3c8,dacColor);
		while(dacSubColor--) outportb(0x3c9,*pp2++);
	}

	outportb(0x3c6,capture->palmask);	// restore palette mask

	// kick the VGA into planar mode (mode X) so we can grab all memory

//	goto skip_all;

	outportb(0x3c4,4);
	outportb(0x3c5,capture->seq[4] & 0xF7);		// turn off 'chain four' mode
	outportb(0x3ce,5);
	outportb(0x3cf,capture->graph[5] & 0xF7);	// go to read mode 0

	for(x=0;x<4;x++) {
		outportb(0x3ce,4);
		outportb(0x3cf,x);

		qmovel((ulong *)capture->plane[x],(ulong *)0xA0000,16384);
	}

	/* restore VGA registers back the way they were */

//	outportb(0x3c4,2);
//	outportb(0x3c5,capture->seq[2]);	/* restore sequencer register 4 */
	outportb(0x3c4,4);
	outportb(0x3c5,capture->seq[4]);	/* restore sequencer register 4 */
	outportb(0x3ce,4);
	outportb(0x3cf,capture->graph[4]);
	outportb(0x3ce,5);
	outportb(0x3cf,capture->graph[5]);	/* restore gc register 5 */

	outportb(0x3c4,capture->seqreg & 0x3f);
	outportb(0x3d4,capture->crtcreg);
	outportb(0x3ce,capture->graphreg & 0x1f);

	opt.scr_captured=TRUE;		/* got one! */
	++vgacm.frames;
}

void	Capture_Simple(void) {
	int x;
	uchar *pp;

	if (!(pp = Capture_Allocate_Memory(64768))) {
		++vgacm.framesbad;
		return;
	}

	qmovel((ulong *)pp,(ulong *)0xA0000,16000);
	pp += 64000;

	outportb(0x3c7,0);
	x=768;
//	while(x--) *pp++=(inportb(0x3c9)*85L)/21;
	while(x--) *pp++=expand6to8(inportb(0x3c9));

	opt.scr_captured = TRUE;
	++vgacm.frames;
}

ulong	movechk(void *,void *,ulong);
#pragma aux movechk=	\
	"xor eax,eax",	\
	"sub edi,esi",	\
	"sub edi,4",	\
	"l:",		\
	"mov ebx,[esi]",\
	"add esi,4",	\
	"add eax,ebx",	\
	"dec ecx",	\
	"mov [esi+edi],ebx",\
	"jne l"	parm [edi] [esi] [ecx] value [eax] modify exact [eax ebx ecx esi edi];

int memcmpl(void *,void *,ulong lc);
#pragma aux memcmpl=	\
	"mov eax,1","rep cmpsd","jnz l","xor al,al","l:" parm [esi] [edi] [ecx] value [eax] modify exact [eax esi edi ecx];

void	Capture_RapidFire(void) {
	int x;
	uchar *pp;
	ulong chksum;

	if (!(pp = Capture_Allocate_Memory(64768))) {
		++vgacm.framesbad;
		vgacm.rapidfire = FALSE;
		return;
	}

	chksum = movechk((ulong *)pp,(ulong *)0xA0000,16000);

	if (chksum == vgacm.last_chk) {
		if (vgacm.last_ptr)
			if (!memcmpl(pp,vgacm.last_ptr,16000)) {
		++vgacm.framesbad;
				Capture_Deallocate_Memory(64768);
				return;
			}
	}

	vgacm.last_chk = chksum;
	vgacm.last_ptr = pp;
	pp += 64000;

	outportb(0x3c7,0);
	x=768;
//	while(x--) *pp++=(inportb(0x3c9)*85L)/21;
	while(x--) *pp++=expand6to8(inportb(0x3c9));

	opt.scr_captured = TRUE;
	vgacm.rapidfire = TRUE;
	++vgacm.frames;
}

void	Capture_RapidFireX(void) {
	int x;
	uchar *pp;
	ulong chksum;
	uchar	graph_reg;
	uchar	graph_bank;
	uchar	c;

	if (!(pp = Capture_Allocate_Memory(64768))) {
		++vgacm.framesbad;
		vgacm.rapidfire = FALSE;
		return;
	}

	graph_reg = inp(0x3ce);
	outp(0x3ce,4);
	graph_bank = inp(0x3cf);

	outp(0x3ce,4); outp(0x3cf,0); chksum  = movechk((ulong *)pp        ,(ulong *)0xA0000,4000);
	outp(0x3ce,4); outp(0x3cf,1); chksum += movechk((ulong *)pp +  4000,(ulong *)0xA0000,4000);
	outp(0x3ce,4); outp(0x3cf,2); chksum += movechk((ulong *)pp +  8000,(ulong *)0xA0000,4000);
	outp(0x3ce,4); outp(0x3cf,3); chksum += movechk((ulong *)pp + 12000,(ulong *)0xA0000,4000);

	outp(0x3ce,4);
	outp(0x3cf,graph_bank);
	outp(0x3ce,graph_reg);

	if (chksum == vgacm.last_chk) {
		if (vgacm.last_ptr)
			if (!memcmpl(pp,vgacm.last_ptr,16000)) {
		++vgacm.framesbad;
				Capture_Deallocate_Memory(64768);
				return;
			}
	}

	vgacm.last_chk = chksum;
	vgacm.last_ptr = pp;
	pp += 64000;

	outportb(0x3c7,0);
	x=768;
//	while(x--) *pp++=(inportb(0x3c9)*85L)/21;
	while(x--) {
		c = inp(0x3c9);
		*pp++ = (c<<2) | (c>>4);
	}

	opt.scr_captured = TRUE;
	vgacm.rapidfire = TRUE;
	++vgacm.frames;
}

#define	LC_LITERAL	(0)
#define	LC_REPEAT	(1)
#define	LC_NONE		(2)

void	Capture_RapidFire_Compressed(void) {
	int i,cnt;
	uchar *pp,*pp0,*vm = (uchar *)0xa0000;
	uchar *key;
	uchar x,y;
	uchar chksum=0;

	char mode=LC_REPEAT;

	if (!(pp = Capture_Allocate_Memory(64500+768))) {
		++vgacm.framesbad;
		vgacm.rapidfire = FALSE;
		return;
	}
	pp0 = pp;

	i = 64000;
	y=*vm++;
	--i;
	while(i) {
		if (mode==LC_REPEAT) {
			x=y;
			y=*vm++;
			if (!--i) {
				mode=LC_LITERAL;
				break;
			};
		};
		if (x==y) {
			mode=LC_REPEAT;
			cnt=2;
			key=pp++;
			*pp++=x;		chksum ^= x;
			for(;;) {
				y=*vm++;
				if (!--i) {
					if (x==y && cnt<128) ++cnt;
					else {
						*pp++=0;
						*pp++=y;	chksum ^= y;
					};
					mode=LC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x!=y) break;
				++cnt;
			};
			*key=1-cnt;
		} else {
			mode=LC_LITERAL;
			cnt=1;
			key=pp++;
			for(;;) {
				*pp++=x;	chksum ^= x;
				x=y;
				y=*vm++;
				if (!--i) {
					if (cnt<127) cnt+=2;
					else {
						*pp++=1;	chksum ^= y;
					};
					*pp++=x;
					*pp++=y;
					mode=LC_NONE;
					break;
				};
				if (cnt==128) break;
				if (x==y) {

					/* if we have ??xx?? include xx in literal run */

					if (*vm==x)	break;
				};
				++cnt;
			};
			*key=cnt-1;
		};
	};

	if (mode==LC_LITERAL) {
		if (x==y) {
			*pp++=0xFF;		/* repeat of 2 */	chksum ^= 0xff;
		} else {
			*pp++=0x01;		/* literal 2 */		chksum ^= 1;
			*pp++=x;
		};
		*pp++=y;
	} else if (mode==LC_REPEAT) {
		*pp++=0x00;	/* literal 1 */
		*pp++=y;
	};

	i = (4-(pp-pp0)) & 3;
	while(i--) *pp++ = 0;

	if (chksum == vgacm.last_chk) {
		if (vgacm.last_ptr)
			if (!memcmpl(pp0,vgacm.last_ptr,(pp-pp0)>>2)) {
				++vgacm.framesbad;
				Capture_Deallocate_Memory(64500+768);
				return;
			}
	}

	vgacm.last_chk = chksum;
	vgacm.last_ptr = pp0;

	outportb(0x3c7,0);
	i=768;
//	while(i--) *pp++=(inportb(0x3c9)*85L)/21;
	while(i--) *pp++=expand6to8(inportb(0x3c9));

	// free off excess memory

	Capture_Deallocate_Memory((64500+768)-((pp+3-pp0)&0xfffffffc));

	opt.scr_captured = TRUE;
	vgacm.rapidfire = TRUE;
	++vgacm.frames;
}

//#pragma aux PrtSc_Interrupt loadds far;
//void __far __loadds __interrupt PrtSc_Interrupt(void) {
void DoCapture(void) {

	Click();

	switch(vgacm.type) {
	case	VCT_SIMPLE:	Capture_Simple();	break;
	case	VCT_RF:		Capture_RapidFire();	break;
	case	VCT_RFX:	Capture_RapidFireX();	break;
	case	VCT_RFC:	Capture_RapidFire_Compressed();	break;
	case	VCT_ADVANCED:	Capture_Advanced();	break;
	case	VCT_VESA:	Capture_VESA();		break;
	}
}

void Capture_VESA(void) {
// real cheap capture of a 640x480x256 screen

	ulong *pm=(ulong *)picmem;
	ulong *vm;
	short b,x;
	static union REGS regs,rregs;

	if (picbm->size<(640*480)) return;

	for(b=0;b<5;b++) {
		regs.w.ax=0x4f05;
		regs.w.bx=0x0000;
		regs.w.dx=b;
		int386(0x10,&regs,&rregs);
		vm=(ulong *)0xA0000;
		x=16384;
		if (b==4) x=11264;
		for(;x;x--) *pm++=*vm++;
	};

	outportb(0x3c7,0);
	{
		uchar *pal=piccol;
		short x=768;
		do {
//			*pal++=(inportb(0x3c9)*85L)/21;
			*pal++=expand6to8(inportb(0x3c9));
		} while(--x);
//		inportbs(0x3c9,piccol,768);
	}
}


long	Capture_Int(void);

struct IHandler Capture_Int08={
	{ {NULL},(IntFuncPtr)&Capture_Int,NULL },
	{ NULL },
	0,
	0,
};

struct IHandler Capture_Int09={
	{ {NULL},(IntFuncPtr)&Capture_Int,NULL },
	{ NULL },
	0,
	0,
};

long	Capture_Int(void) {
	uchar	joyread;

	if (vgacm.rapidfire) DoCapture();
	else if (vgacm.joyMask) {
		joyread = JoyRead | JoyRead | JoyRead;

		if (vgacm.joyMask != ((joyread | ~vgacm.lastJoyRead) & vgacm.joyMask)) DoCapture();

		vgacm.lastJoyRead	= joyread;
	}
	return 0;
}


