#define VGAPAINT_DITHER_C

#include <a/bits.h>

#include "vgapaint.h"
#include "mouse.h"
#include "picture.h"

void Dither2(void) {
	short	*even_row,		/* even row memory */
		*odd_row,		/* odd row memory */
		*trow,			/* pointer to this row's element */
		*nrow;			/* pointer to element below and to left */
	uchar	pmap[256];		/* palette mapped to brightness */
	short x;

	if (!(even_row=malloc(picbm->width*4+4))) return;
	if (!(odd_row=malloc(picbm->width*4+4))) {
		free(even_row);
		return;
	};

	mouse_suspend();
	vd->init(VD_TO13,0,0);
	vid_setmode(0x12);	/* 640x480x2 colors */
	vid_setRGB6(0,227,227,227);
	vid_setRGB6(1,0,0,0);
	vid_setpreg(0,0);
	vid_setpreg(15,1);

/* first order of business... compute each color value down to a
** single value 0-255 using 30%-59%-11% ratio for RGB conversion.
*/
	{
		short bright;
		uchar *map=pmap,*colsrc=piccol;

		for(x=0;x<256;x++) {
			bright=30L*(*colsrc++);
			bright+=(59L*(*colsrc++));
			bright+=(11L*(*colsrc++));
			*map++=bright/100;
		};
	}

// Preform Floyd-Steinberg (F-S) dithering on the picture.
//
// For each pixel, plot its point according to the current error value
// (should be 128).  Then compute the error between the desired pixel and
// the plotted pixel.  Distribute the error as follows:
//
//	 ... here 7/16
//	3/16 5/16 1/16
//
// Work from left-to-right on even rows and right-to-left on odd rows.

	qsetw(odd_row,0,picbm->width*2+2);

	{
		uchar *srcmem=picmem;
		uchar *vidmem=(uchar *)0xA0000;
		short x2,y,error;
		uchar byte,mask;
		short pxl,pxl2;
		boolean last;
		long	direction=1;
		char	rotdir=1;
		uchar	imask=0x80;

		for (y=400;y;y--) {
			swap(short *,even_row,odd_row);
			qsetw(odd_row,0,picbm->width*2+2);
			trow=even_row+1;
			nrow=odd_row+641;

			for (x=80;x;x--) {
				mask=imask;
				byte=0;
				for (x2=4;x2;x2--) {
					pxl=pmap[*srcmem];
					srcmem+=direction;

					last=FALSE;
					do {
						pxl2=pxl+((*trow++ +8)>>4);
						if (pxl2<128) {
							byte+=mask;
							error=pxl2;
						} else	error=pxl2-255;

						nrow[-2] += error;
						nrow[ 0] += error*3;
						nrow[-1] += error*5;
						*trow    += error*7;

						--nrow;
						mask=brotr(mask,rotdir);
					} while(last=!last);
				};
				*vidmem=byte;
				vidmem+=direction;
			};

			if (y&1) {
				srcmem += (vidbm.bpr + 1);
				vidmem += 81;
			} else {
				vidmem += 79;
				--srcmem;
			};

			direction=-direction;
			rotdir=8-rotdir;
			imask=0x81-imask;
		};
	}

//	mouse_wait_key();
	while(!mouse_buttons() && !kbhit()) DPMI_Yield();

	vid_setmode(0x13);	/* 320x200x256 colors */
	vd->init(VD_FROM13,0,0);
	vid_loadRGB6(0,piccol,256);
	restore_screen();
	mouse_resume();

	free(even_row);
	free(odd_row);
}

uchar col16[]={
	255,255,255,
	 85,255,255,
	255, 85,255,
	 85, 85,255,
	255,255, 85,
	 85,255, 85,
	255, 85, 85,
	 85, 85, 85,
	170,170,170,
	  0,170,170,
	170,  0,170,
	  0,  0,170,
	170,170,  0,
	  0,170,  0,
	170,  0,  0,
	  0,  0,  0,
};

typedef	short *DTABLEPTR;

void Dither16(void) {
	DTABLEPTR even_row=NULL;
	DTABLEPTR odd_row=NULL;
	short x,p;
	boolean abort=FALSE;
	uchar pal[1024];
	uchar *dst=pal;
	uchar *src=piccol;

	if (
		!(even_row=(DTABLEPTR)malloc(picbm->width*4+4)) ||
		!(odd_row=(DTABLEPTR)malloc(picbm->width*4+4))) {
		abort=TRUE;
		goto xit;
	};

	for(x=256;x;x--) {
#if 0
		*dst++=expand6to8(*src++);
		*dst++=expand6to8(*src++);
		*dst++=expand6to8(*src++);
#else
		*dst++=*src++;
		*dst++=*src++;
		*dst++=*src++;
#endif
		*dst++=((long)(30*src[-3]+59*src[-2]+11*src[-1])<<8)/25501;
	};

/* now dither the picture, using four pixel groups for each source
** pixel...
*/

	mouse_suspend();	/* shut down pointer */
	vid_setmode(0x11);	/* 640x480x16 colors */
	vid_loadRGB6(0,col16,16);
	for (x=0;x<16;x++)
		vid_setpreg(x,x);

	outportb(0x3ce,0x05);	/* VGA register 5 (mode) */
	outportb(0x3cf,0x00);	/* write mode 0 */
	outportb(0x3ce,0x08);	/* VGA register 1 (bitmask) */
	outportb(0x3cf,0xff);	/* all bits */

	src=pal;
	for(p=1;p<16;p<<=1) {
		uchar *srcmem=picmem;
		uchar *vidmem=(uchar *)0xA0000;
		short *trow,*nrow;
		short x2,y,error;
		uchar byte,mask;
		short pxl,pxl2;
		boolean last;
		long	direction=1;
		char	rotdir=1;
		uchar	imask=0x80;

		outportb(0x3c4,0x02);	/* Sequencer register 1 (map mask) */
		outportb(0x3c5,p);	/* set the plane to modify */

		qsetw(odd_row,0,picbm->width*2+2);

		for (y=400;y;y--) {
			swap(short *,even_row,odd_row);
			qsetw(odd_row,0,picbm->width*2+2);
			trow=even_row+1;
			nrow=odd_row+641;

			for (x=80;x;x--) {
				mask=imask;
				byte=0;
				for (x2=4;x2;x2--) {
					pxl=src[*srcmem*4];
					srcmem+=direction;

					last=FALSE;
					do {
						pxl2=pxl+((*trow++ +8)>>4);
						if (pxl2<128) {
							byte+=mask;
							error=pxl2;
						} else	error=pxl2-255;

						nrow[-2] += error;
						nrow[ 0] += error*3;
						nrow[-1] += error*5;
						*trow    += error*7;

						--nrow;
						mask=brotr(mask,rotdir);
					} while(last=!last);
				};
				*vidmem=byte;
				vidmem+=direction;
			};

			if (y&1) {
				srcmem+=vidbm.bpr + 1;
				vidmem+=81;
			} else {
				vidmem+=79;
				--srcmem;
			};

			direction=-direction;
			rotdir=8-rotdir;
			imask=0x81-imask;
		};
		++src;
	}

	outportb(0x3ce,0);
	outportb(0x3cf,0);
	outportb(0x3ce,1);
	outportb(0x3cf,0);
	outportb(0x3ce,8);
	outportb(0x3cf,0xff);

	while(!mouse_buttons() && !kbhit()) DPMI_Yield();
//	mouse_wait_key();

xit:

	if (even_row) free(even_row);
	if (odd_row)  free(odd_row);

	vid_setmode(0x13);	/* 320x200x256 colors */
	vid_loadRGB6(0,piccol,256);
	restore_screen();
	mouse_resume();

	if (abort)
		Message("Not enough memory!");

}

///////////////////////////////////////////////////////////////////////

boolean	Dither8_function(struct Picture *pict) {
	uchar		*palette	= pict->current->frame_pal ? pict->current->frame_pal->palette : pict->palette;
	DTABLEPTR	row_mem		= NULL;
	DTABLEPTR	even_row	= NULL;
	DTABLEPTR	odd_row		= NULL;
	uchar		*dst		= pict->bm->vm;
	short		*trow,*nrow;
	int		x,y,err,pxl,pxl2;
	uchar		c;
	long		direction=1;

	if (!(row_mem = even_row = (DTABLEPTR)malloc(sizeof(short)*(picbm->width+2)*6))) return FALSE;
	odd_row = even_row + (picbm->width+2)*3;


	qsetw(odd_row,0,pict->bm->width*3+6);

	y = pict->bm->height;
	do {
		swap(short *,even_row,odd_row);
		qsetw(odd_row,0,picbm->width*3+6);
		trow	= even_row + 3;
		nrow	= odd_row  + pict->bm->width*3;

		x = pict->bm->width;
		do {
			c=0;

			err=palette[3**dst + 0] + ((trow[0]+8)>>4);
			if (err>=128)	{ c += 4; err -= 255; }
			nrow[-3] += err;
			nrow[ 3] += err*3;
			nrow[ 0] += err*5;
			trow[ 3] += err*7;

			err=palette[3**dst + 1] + ((trow[1]+8)>>4);
			if (err>=128)	{ c += 2; err -= 255; }
			nrow[-2] += err;
			nrow[ 4] += err*3;
			nrow[ 1] += err*5;
			trow[ 4] += err*7;

			err=palette[3**dst + 2] + ((trow[2]+8)>>4);
			if (err>=128)	{ c += 1; err -= 255; }
			nrow[-1] += err;
			nrow[ 5] += err*3;
			nrow[ 2] += err*5;
			trow[ 5] += err*7;

			*dst = c;
			dst+=direction;
			nrow-=3;
			trow+=3;
		} while(--x);

		if (direction>0)
			dst += pict->bm->bpr - 1;
		else
			dst += pict->bm->bpr + 1;

		direction=-direction;
	} while(--y);

	free(row_mem);

	DetachFramePalette(pict,pict->current);

	return TRUE;
}

boolean	Dither8(struct Picture *pict) {
	int i;
	uchar *p = pict->palette;

	PictureProcessFrames(pict,Dither8_function,"Dithering picture to 8 colors");

	for(i=0;i<8;i++) {
		p[0] = i&4?0xff:0x00;
		p[1] = i&2?0xff:0x00;
		p[2] = i&1?0xff:0x00;
		p+=3;
	}
	pict->pal_bits[0]=0xff;
	memset(pict->pal_bits+1,0,31);

	return TRUE;
}

///////////////////////////////////////////////////////////////////////

void colorbars(void) {
	short x,y;
	uchar c=0;
	vd->set(picbm,0);

	for (y=0;y<192;y+=12)
		for(x=0;x<320;x+=20)
			vd->bar(picbm,x,y,20,12,c++);

	rectmodified=FALSE;	/* can't undo this */

	restore_screen();
}

void colorbars2(void) {
	short c,c2=0;
	short y;
	uchar *col=piccol;

	vd->set(picbm,0);
	for (c=0;c<64;c++) {
		vd->bar(picbm,c2,  0,4,50,c);
		vd->bar(picbm,c2, 50,4,50,c+64);
		vd->bar(picbm,c2,100,4,50,c+128);
		vd->bar(picbm,c2,150,4,50,c+192);
		c2+=4;
		y=(c*85)/21;
		col[  0]=y;
		col[  1]=0;
		col[  2]=0;
		col[192]=0;
		col[193]=y;
		col[194]=0;
		col[384]=0;
		col[385]=0;
		col[386]=y;
		col[576]=y;
		col[577]=y;
		col[578]=y;
		col+=3;
	};

	rectmodified=FALSE;

	repick_colors();
	restore_screen();
}

static uchar rainbow_pal[]={
	255,  0,  0,
	255,  0, 93,
	255,  0,194,
	227,  0,255,
	125,  0,255,
	 32,  0,255,
	  0, 61,255,
	  0,158,255,
	  0,255,255,
	  0,255,162,
	  0,255, 61,
	 28,255,  0,
	130,255,  0,
	223,255,  0,
	255,190,  0,
	255, 97,  0,
};

void prescr_rainbows(void) { 
	uchar *pc=piccol,*pc2=piccol+14*16*3,*rc;
	short x,y,wht;
	uchar r,g,b;

	for(x=1;x<=8;x++) {
		rc=rainbow_pal;
		wht=(8-x)*255+4;
		for(y=16;y;y--) {
			r=*rc++;
			g=*rc++;
			b=*rc++;
			*pc ++=((short)r*x+  4)>>3;
			*pc2++=((short)r*x+wht)>>3;
			*pc ++=((short)g*x+  4)>>3;
			*pc2++=((short)g*x+wht)>>3;
			*pc ++=((short)b*x+  4)>>3;
			*pc2++=((short)b*x+wht)>>3;
		};
		pc2-=96;
	};

	pc=piccol+15*16*3;
	for(x=0;x<16;x++) {
		r=((short)x*21)/5;
		*pc++=r;
		*pc++=r;
		*pc++=r;
	};

	vid_loadRGB6(0,piccol,256);
	vd->set(picbm,0);

	r=0;

	for (y=0;y<192;y+=12)
		for(x=0;x<320;x+=20)
			vd->bar(picbm,x,y,20,12,r++);

	rectmodified=FALSE;	/* can't undo this */

	repick_colors();
	restore_screen();
}

