#define	VGAPAINT_MOUSE

#include "vgapaint.h"
#include "mouse.h"
#include "io.h"

// Linux DOSEMU does not seem to obey 0x33 interrupt calls given from
// protected mode

#define MOUSE_NO_INT386

#ifdef MOUSE_NO_INT386
typedef struct RealInt REGSTRUCT;
#define INT386(i,regs) DPMIRealInt((i),(regs));
#else
typedef union REGS REGSTRUCT;
#define INT386(i,regs) int386((i),(regs),(regs));
#endif

// mouse routines are called frequently, turn off stack checking

#pragma off (check_stack)


void	mouse_bounds_reset(void) {
//	mouse_setlimits(-128,-64,(vidbm.width<<1)+128-1,vidbm.height+64-1);
	mouse_setlimits(-ms_pan*2,-ms_pan,(vidbm.width<<1)+ms_pan*2-1,vidbm.height+ms_pan-1);
	mouse_move(vidbm.width>>1,vidbm.height>>1);
	mouse_ptrinit();
	mouse_setpointer(crosshairs,7,7,7,0);
	mouse_set_split(0,0);
	ms_hideflag=1;
	mouse_show();
};

extern uchar far *mouse_stack_pointer;

ulong	mouse_stack[1024];

int mouse_init(void) {
	REGSTRUCT regs;
	int btns;

	mouse_stack_pointer = mouse_stack+1024;

	// initialize mouse structures

	memset(&ms_mpi,0,sizeof ms_mpi);
	ms_mpi.high=7;

	// if we are running under Windows 95, adjust the bias as far as
	// it will go (16k) so we can detect windowed mode

	if (ms_bias && sys.Win95) {
		ms_bias = 16384;
		ms_mpi.detect_window = TRUE;
	}

	// initialize mouse driver

	if (!*(long *)(0x33 << 2)) return -1; /* no mouse driver */
	regs.x.eax=NULL;
	INT386(0x33,&regs);
	if (!regs.x.eax) return -2; /* no mouse connected */
	btns=regs.w.bx;

	regs.w.ax=2;	/* hide ptr */
	INT386(0x33,&regs);

	return btns;		/* return number of buttons */
}

// make sure we don't cause sluggish mouse problems

void mouse_show(void) {
	if (ms_hideflag)
		if (!--ms_hideflag) {
			asm_mouse_disable();
			mouse_draw_ptr(ms_msi.scr_x-ms_mpi.hot_x,ms_msi.scr_y-ms_mpi.hot_y);
			asm_mouse_enable();
		};
}

void mouse_hide(void) {
	asm_mouse_disable();
	++ms_hideflag;
	mouse_erase_ptr();
	asm_mouse_enable();
}

void mouse_setpointer(uchar *cursdata,mouse_coord xhotspot,mouse_coord yhotspot,color high,color low) {
	ulong	xaccum,yaccum,xinc,yinc;
	uchar	*dp,*dp2,*sp;

	if (ms_mpi.dynamic_image) {
		free(ms_mpi.image);
		ms_mpi.dynamic_image=0;
	}

	// allocate enough room for the new image and zoom it

	if (!(ms_mpi.image = malloc(((16*XPOINT+3)&~3L)*16*YPOINT)))
		goto abort;

	ms_mpi.dynamic_image = 2;

	xinc = (0x0FFFFL+XPOINT)/XPOINT;
	yinc = (0x0FFFFL+YPOINT)/YPOINT;

	xaccum = yaccum = 0L;

	dp2 = ms_mpi.image;

	do {
		sp = cursdata + 16*(yaccum>>16);
		dp = dp2;	dp2 += (16*XPOINT+3)&~3L;
		xaccum=0;
		do {
			*dp++ = sp[(xaccum>>16)];
			xaccum += xinc;
		} while(xaccum<0x100000L);
		yaccum+=yinc;
	} while(yaccum<0x100000L);

	if (!mouse_setpointer2(ms_mpi.image,xhotspot*XPOINT+(XPOINT>>1),yhotspot*YPOINT+(YPOINT>>1),(16*XPOINT+3)&~3L,16*YPOINT,high,low))
		goto abort;

	return;
abort:
	Message("Warning: Not enough memory for mouse pointer.  Save your work and exit.");
	asm_mouse_disable();
	free(ms_mpi.image);
}

boolean	mouse_setpointer2(uchar *cursdata,mouse_coord xhot,mouse_coord yhot,mouse_coord width,mouse_coord height,color high,color low) {

	if (vd->changeptr)
		asm_mouse_disable();
	else
		mouse_hide();

	if (ms_mpi.dynamic_image==1) {
		free(ms_mpi.image);
		ms_mpi.dynamic_image=0;
	} else if (ms_mpi.dynamic_image==2) {
		ms_mpi.dynamic_image=1;
	}

	// allocate memory for the backup and possibly for image blit memory in case
	// the driver is lazy (like the VESA driver)

	if (ms_mpi.s_width*ms_mpi.s_height != width*height)
		if (!(ms_mpi.backup = realloc(ms_mpi.backup,width*height*2))) return FALSE;

	ms_mpi.blitmem = ms_mpi.backup + width*height;

	ms_mpi.s_width = width;
	ms_mpi.s_height = height;

	ms_mpi.image	= cursdata;
	ms_mpi.changed	= TRUE;
	ms_mpi.hot_x	= xhot;
	ms_mpi.hot_y	= yhot;
	ms_mpi.high	= high;
	ms_mpi.low  	= low;

	if (vd->changeptr) {
		vd->changeptr(&ms_mpi);
		asm_mouse_enable();
	} else
		mouse_show();

	return TRUE;
}

int	mouse_buttons_direct(void) {
	REGSTRUCT regs;

	regs.x.eax=0x0003;
	INT386(0x33,&regs);
	return regs.w.bx;
}

int	mouse_buttons(void) {
	return ms_msi.buttons;
}

void mouse_read(mouse_coord *xptr,mouse_coord *yptr,mouse_coord *button) {
	boolean shiftit=FALSE;
	long sx,sy,px,py;
	REGSTRUCT regs;

kbhit();	// admittedly a lame way to check for break

//	int386(0x28,&regs,&regs);	/* DosOK interrupt */
//	regs.x.eax=0x0003;

//	int386(0x33,&regs,&regs);

//	sx=(long)regs.x.ecx-ms_bias;	/* correct for bias */
//	sy=(long)regs.x.edx-ms_bias;

	px	= ms_msi.pos_x<<1;
	py	= ms_msi.pos_y;
	sx	= ms_msi.scr_x<<1;
	sy	= ms_msi.scr_y;

	if (px<0) {
		ms_xscroll+=px;
		px=0;
		shiftit=TRUE;
	} else if (px>=(vidbm.width<<1)) {
		ms_xscroll+=px-(vidbm.width<<1)+1;
		px=(vidbm.width<<1)-1;
		shiftit=TRUE;
	};
	if (py<0) {
		ms_yscroll+=py;
		py=0;
		shiftit=TRUE;
	} else if (py>=vidbm.height) {
		ms_yscroll+=py-vidbm.height+1;
		py=vidbm.height-1;
		shiftit=TRUE;
	};

	*xptr=sx;
	*yptr=sy;
//	*button=regs.x.ebx;
	*button = ms_msi.buttons;

	if (shiftit) {
		ms_msi.pos_x=px>>1;
		ms_msi.pos_y=py;
		regs.x.eax=0x0004;
		regs.x.ecx=(long)(ms_bias+px);
		regs.x.edx=(long)(ms_bias+py);
		INT386(0x33,&regs);
	};

//	if (*button & 4) {
//		if (!restore_scrn) {
//			PrtSc_Interrupt2();
//			restore_scrn=TRUE;
//		};
//	};
}

void mouse_read2(mouse_coord *xptr,mouse_coord *yptr,mouse_coord *button) {
	mouse_read(xptr,yptr,button);
	*xptr>>=1;
}

void mouse_readmove(mouse_coord *x,mouse_coord *y,mouse_coord *b) {
	mouse_coord X,Y,B;
	do {
		mouse_read2(&X,&Y,&B);
		CheckKeyInterrupt();
		if (mouse_readmoveflag()) break;		
		DPMI_Yield();
	} while((*b == B) && !dabort);
	*x=X;
	*y=Y;
	*b=B;
}

void mouse_wait_key(void) {
	mouse_coord x,y,b;
	vd->refresh();
	do { kbhit(); mouse_read2(&x,&y,&b); } while(!b);
}

void mouse_wait_unkey(void) {
	mouse_coord x,y,b;
	vd->refresh();
	do { kbhit(); mouse_read2(&x,&y,&b); } while (b);
}


/****** ATTENTION: This is HARDCODED into mouse_read()! ******/

/* the mouse driver can't take negative values... so add ms_bias */

void mouse_setlimits(mouse_coord x1,mouse_coord y1,mouse_coord x2,mouse_coord y2) {
	REGSTRUCT regs;

	regs.x.eax=0x0007;
	regs.x.ecx=(long)(x1+ms_bias);
	regs.x.edx=(long)(x2+ms_bias);
	INT386(0x33,&regs);
	regs.x.eax=0x0008;
	regs.x.ecx=(long)(y1+ms_bias);
	regs.x.edx=(long)(y2+ms_bias);
	INT386(0x33,&regs);
	ms_savelimits.x1=x1;
	ms_savelimits.y1=y1;
	ms_savelimits.x2=x2;
	ms_savelimits.y2=y2;
}

// mouse_readlimits: read the current mouse limits

void	mouse_readlimits(MouseBound *mbs) {
	*mbs=ms_savelimits;
} 

// mouse_writelimits: write the current mouse limits

void	mouse_writelimits(MouseBound *mbs) {
	mouse_setlimits(mbs->x1,mbs->y1,mbs->x2,mbs->y2);
}

/* mouse_suspend: suspend the mouse's operation */

void mouse_suspend(void) {
	mouse_hide();
	mouse_ptrdispose();
}

/* mouse_resume: restore mouse operation after a suspend */

void mouse_resume(void) {
	mouse_ptrinit();
	mouse_show();
	mouse_setlimits(ms_savelimits.x1,ms_savelimits.y1,ms_savelimits.x2,ms_savelimits.y2);
}

void mouse_move(mouse_coord x,mouse_coord y) {
	REGSTRUCT regs;

	ms_msi.scr_x = x;
	ms_msi.scr_y = y;

	if (y < ms_mpi.split_2)
		y += ms_mpi.split_1;
	else
		y -= ms_mpi.split_2;

	ms_msi.pos_x = x;
	ms_msi.pos_y = y;
	regs.x.eax=0x0004;
	regs.x.ecx=ms_bias+(x<<1);
	regs.x.edx=ms_bias+y;
	INT386(0x33,&regs);
}

// mouse_set_split: set the "split" level of the mouse cursor
//
//	y1: position of split
//	y2: height of split section

void mouse_set_split(mouse_coord y1,mouse_coord y2) {
	ms_mpi.split_1 = y1;
	ms_mpi.split_2 = y2;
}



void mouse_ptrinit(void) {
	if (!ms_ptrinst) {
		if (!sys.realDPMI) {
			union REGS regs;
			struct SREGS sregs;

	// if no "real" DPMI server is running, we must be using DOS/4GW
	// therefore, we can hook using its own mouse handler
	// **WARNING** we must use int386 for this, not DPMIRealInt

			segread(&sregs);

			regs.w.ax=0x0014;
			regs.w.cx=0x007F;
			regs.x.edx=FP_OFF((void far *)&asm_mouse_ptrfunc);
			sregs.es=FP_SEG((void far *)&asm_mouse_ptrfunc);
			int386x(0x33,&regs,&regs,&sregs);
			ms_oldmask=regs.w.cx;
			ms_oldoff=regs.x.edx;
			ms_oldseg=sregs.es;
		} else {
			struct RealInt RMI;

	// We are running either under PMODE/W, or a real DPMI server
	// to be safe with Windows 3.1, install a real mode callback since
	// we have them (Win3.1 has several nasty bugs in its 32-bit
	// DOS emulation code)

			ms_cback=DPMI_AllocCBack(&asm_mouse_ptrf2,ms_cbackbuf);
			if (!ms_cback) exit(5);

			RMI.w.ax=0x0014;
			RMI.w.cx=0x007F;

			RMI.es=ms_cback>>16;
			RMI.w.dx=ms_cback;

			DPMIRealInt(0x33,&RMI);

			ms_oldmask=regs.w.cx;
			ms_oldoff=regs.w.dx;
			ms_oldseg=sregs.es;
		};

		ms_ptrinst=TRUE;		
	};
}

void mouse_ptrdispose(void) {
	if (ms_ptrinst) {
		if (!sys.realDPMI) {
			struct SREGS sregs;

			segread(&sregs);

			regs.w.ax=0x0014;
			regs.w.cx=ms_oldmask;
			regs.x.edx=ms_oldoff;
			sregs.es=ms_oldseg;
			int386x(0x33,&regs,&regs,&sregs);
		} else {
			struct RealInt RMI;

			RMI.w.ax=0x0014;
			RMI.w.cx=ms_oldmask;
			RMI.w.dx=ms_oldoff;
			RMI.es=ms_oldseg;
			DPMIRealInt(0x33,&RMI);

			DPMI_FreeCBack(ms_cback);
		};
		ms_ptrinst=FALSE;
	};
}


/* Add ALL normal VGAPaint routines BEFORE this point */


#pragma off (check_stack)

void mouse_pointerfunc(ushort cmask,ushort bstate,short choriz,short cvert,short hcnts,short vcnts) {
#pragma aux mouse_pointerfunc parm [EAX] [EBX] [ECX] [EDX] [EDI] [ESI];

//	if (ms_hideflag) return;

	// check for windowed/full screen mode?

	if (ms_mpi.detect_window) {
		boolean	newwinstate;

		// if the x-coord is below 12,288 assume a window

		newwinstate = (choriz < 12288);

		if (newwinstate ^ ms_mpi.in_window) {
			ms_mpi.in_window = newwinstate;
			ms_mpi.winstate_chgd = TRUE;
		}

	}

	if (!ms_mpi.detect_window || !ms_mpi.in_window) {
		choriz-=ms_bias;		/* correct for bias */
		cvert-=ms_bias;
	}

	choriz>>=1;

	ms_msi.pos_x   = choriz;
	ms_msi.pos_y   = cvert;
	ms_msi.buttons = bstate;

	if (choriz<0) choriz=0; else if (choriz>=vidbm.width ) choriz=vidbm.width -1;
	if (cvert <0) cvert =0; else if (cvert >=vidbm.height) cvert =vidbm.height-1;

	if (cvert >= ms_mpi.split_1)
		cvert -= ms_mpi.split_1;
	else
		cvert += ms_mpi.split_2;

	ms_mpi.cur_x = ms_msi.scr_x = choriz;
	ms_mpi.cur_y = ms_msi.scr_y = cvert;

	if (ms_oldx != choriz || ms_oldy != cvert) {
		ms_moved=TRUE;
		ms_oldx=choriz;
		ms_oldy=cvert;
	};

	choriz	-= ms_mpi.hot_x;
	cvert	-= ms_mpi.hot_y;

	if (!ms_hideflag)
		mouse_draw_ptr(choriz,cvert);
}

void mouse_erase_ptr(void) {
	if (!ms_mpi.drawn) return;

	vd->eraseptr(&ms_mpi);

	ms_mpi.drawn=FALSE;
}

void mouse_draw_ptr(mouse_coord x,mouse_coord y) {
	if (ms_mpi.drawn) mouse_erase_ptr();

	if (x>=vidbm.width) return;
	if (y>=vidbm.height) return;
	if (x<=-ms_mpi.s_width) return;
	if (y<=-ms_mpi.s_height) return;

	ms_mpi.last_x = x;
	ms_mpi.last_y = y;
	vd->drawptr(&ms_mpi,x,y);

	ms_mpi.drawn=TRUE;
}


