//////////////////////////////////////////////////////////////
//
// VGAPaint 386 - picload.c
//
// Routines for automatically determining a file's type.
//

#include "picload.h"
#include "io.h"
#include "pcx.h"
#include "list.h"

#define	ARRAYSIZ(x)	(sizeof x/sizeof x[0])

int IFFAutoLoader(char *f) { IFFRead(f); return PIOERR_NONE; }
int BMPAutoLoader(char *f) { BMPRead(f); return PIOERR_NONE; }
int PCXAutoLoader(char *f) { PCXRead(f,2); return PIOERR_NONE; }
int PAKAutoLoader(char *f) { Raw256_Extract(f); return PIOERR_NONE; }

static struct PictureAutoLoader builtin_autoload_list[]={
	{{NULL},   0,PCAL_MATCH,   1,"\xa"	,"pcx",PCXAutoLoader	},
	{{NULL},   0,PCAL_MATCH,   4,"FORM"	,"lbm",IFFAutoLoader	},
	{{NULL},   0,PCAL_MATCH,   4,"FORM"	,"ppm",IFFAutoLoader	},
	{{NULL},   0,PCAL_MATCH,   2,"BM"	,"bmp",BMPAutoLoader	},
	{{NULL},-128,PCAL_MATCH,   0,NULL	,NULL ,PAKAutoLoader	},	// do not give this an extension
};

static struct List PictureAutoloadList;
static int PictureAutoloaderCount=0;

void	InitializeAutoloadSystem(void) {
	struct PictureAutoLoader *pp;
	int i;

	NewList(&PictureAutoloadList);

	pp = builtin_autoload_list;
	for(i=ARRAYSIZ(builtin_autoload_list);i;--i)
		AddTail(&PictureAutoloadList,(struct Node *)pp++);

	PictureAutoloaderCount = ARRAYSIZ(builtin_autoload_list);
}

int	PictureAutoload(char *filename) {
	struct PictureAutoLoader *pp;
	ulong	size=0;
	ulong	fsiz;
	uchar	*buffer;
	FILE	*f;

	// scan the autoload list; find the maximum length

	pp = (struct PictureAutoLoader *)PictureAutoloadList.prev;
	while(pp->node.prev) {
		if (pp->len > size) size = pp->len;
		pp = (struct PictureAutoLoader *)pp->node.prev;
	}

	// allocate a buffer

	if (!(buffer = malloc(size))) return PIOERR_MEMORY;

	// open the file

	if (!(f = VPfopen(filename,"rb"))) {
		free(buffer);
		return PIOERR_OPEN;
	}

	// read the largest chunk we need to analyze

	if ((fsiz=fread(buffer,1,size,f))<=0) {
		free(buffer);
		fclose(f);
		return PIOERR_READ;
	}

	// close the file

	fclose(f);

	// crawl down the match list, looking for a match

	pp = (struct PictureAutoLoader *)PictureAutoloadList.prev;
	while(pp->node.prev) {
		// if the file is shorter than the ID, there's no point in checking

		if (fsiz >= pp->len) {

			// if the autoload type is PCAL_MATCH, try to match the string
			// if the autoload type is PCAL_FUNCTION, call the function to match

			switch(pp->type) {
			case PCAL_MATCH:
				if (!memcmp(buffer,pp->match_address,pp->len))
					goto match;	// ...need double break...
				break;
			case PCAL_FUNCTION:
				if (pp->match_function(buffer,fsiz))
					goto match;	// ...want double break...
				break;
			}
		}

		pp = (struct PictureAutoLoader *)pp->node.prev;
	}

	// this point cannot be reached without a match, since the PAK autoloader will always trigger

match:
	free(buffer);
	return pp->load_function(filename);
}

void	PictureLoad(void) {
	struct PictureAutoLoader *pp;
	char *s,*t;

	if (!(s = (char *)malloc(PictureAutoloaderCount*4)))
		return;

	t=s;
	pp = (struct PictureAutoLoader *)PictureAutoloadList.prev;
	while(pp->node.prev) {
		if (pp->extension) {
			strcpy(t,pp->extension);
			while(*t) ++t;
			*t++ = ';';
		}
		pp=(struct PictureAutoLoader *)pp->node.prev;
	}
	t[-1]=0;

	if (t=GetFileName("Load picture",s))
		if (*t)
			PictureAutoload(t);

	free(s);
}

