#ifndef	VGAPAINT_PICTURE_H
#define	VGAPAINT_PICTURE_H

#include <a/types.h>

#include "ccycle.h"

#define	PICT_HASNAME	(0x00000001L)

enum PalSets {
	PALSET_DEFAULT	= 0,
	PALSET_CCYCLE	= 1,
	PALSET_USER	= 9,
};

struct Picture {
	struct Picture	*next,*prev;	// links to other pictures
	struct BitMap	*bm;		// bitmap describing this picture
	struct Frame	*first;		// pointer to first frame
	struct Frame	*last;		// pointer to last frame
	struct Frame	*current;	// pointer to current frame
	int		current_num;	// current frame number
	int		total_frames;	// # of frames in this anim
	ulong		playback_speed;	// speed to playback at, as 16-bit fraction; 0x08000 is 2/sec
	uchar		*palette;	// palette of this picture
	uchar		*pal_bits;	// palette used bits - determines deleted entries
	uchar		*cur_pal;	// palette of current frame
	uchar		*frame_buf;	// frame buffer used for editing
	uchar		*pack_buf;	// pack buffer used for temporary packing - CAN BE NULL
	ulong		maxframe;	// maximum size of frame (uncompressed size)
	ulong		flags;		// flags
	struct FramePalette *palettes;	// picture's frame specific palettes
	struct PaletteSet *palsets;	// palette sets
	int		current_palset;	// current palset #
	boolean		dirty;		// set if this frame needs to be flushed
	boolean		inited;		// set if this picture exists
	uchar		name[512+2];	// 512 + 2 pad bytes... grab'em!
	struct CycleRange cyclerange[8];
	char		*name_ptr;
};

struct Frame {
	struct Frame	*next,*prev;	// links
	uchar		*frame_mem;	// pointer to frame memory
	struct FramePalette *frame_pal;	// frame's palette - NULL if default used
	ulong		frame_siz;	// size of compressed frame
	boolean		compressed;	// set if this frame is compressed at all
};

struct FramePalette {
	struct FramePalette *next,*prev;
	ulong		usecount;
	uchar		palette[768];
};

struct PaletteSet {
	struct PaletteSet *next,*prev;
	int		length;
	uchar		set[256];
};

boolean	PictureInit(struct Picture *pict, struct BitMap *bm, long width, long height);
void	PictureDispose(struct Picture *pict);
boolean	PictureReinit(struct Picture *pict,long width,long height);
void	PictureSetName(struct Picture *pict,char *name);
boolean	PictureFlush(struct Picture *pict);
void	PictureDecompress(struct Picture *pict);
void	PictureDecompressTo(struct Picture *pict,struct Frame *frame,uchar *buff);
boolean	PictureGotoFrame(struct Picture *pict, struct Frame *frame, int frame_num);
boolean	PictureNextFrame(struct Picture *pict);
boolean	PicturePrevFrame(struct Picture *pict);
void	PictureMarkDirty(struct Picture *pict);
void	PictureInvalidate(struct Picture *pict);
ulong	PredictCompressedFrame(uchar *mem,long width,long height,long bpr);
ulong	CompressFrame(uchar *pack_buf,uchar *frame,long width,long height,long bpr, long maxsize);
void	CopyToFrame(uchar *dest, uchar *src, long width, long height, long bpr);
void	CopyFromFrame(uchar *dest, uchar *src, long width, long height, long bpr);
boolean	InsertBeforeFrame(struct Picture *pict);
boolean	InsertAfterFrame(struct Picture *pict);
boolean	DeleteCurrentFrame(struct Picture *pict);
boolean AllocateFrameBuffer(struct Frame *frame,ulong siz,boolean compressed);
struct FramePalette *CreateFramePalette(void);
void	AttachPaletteToPicture(struct Picture *pict,struct FramePalette *fpal);
void	DetachPaletteFromPicture(struct Picture *pict,struct FramePalette *fpal);
boolean AttachPaletteToFrame(struct Picture *pict,struct Frame *frame,struct FramePalette *pal);
void	DetachFramePalette(struct Picture *pict,struct Frame *frame);
void	LoadFramePalette(struct Picture *pict,struct Frame *frame);
struct Frame *GetFrameAddress(struct Picture *pict,int frame_num);
boolean	PictureProcessFrames(struct Picture *pict,boolean (*func)(struct Picture *pict),char *sb_title);

// from pict_a.asm

extern uchar *Frame_Decompress(uchar *dest,uchar *src,long width,long height,long bpr);

#endif

