// PolygonC.c - Scan conversion of convex polygons

#include <a/types.h>

#include "polygon.h"

struct PolyConvexEdge {
	int	x;		// x-coordinate
	int	idx;		// integer displacement per y change
	int	disc;		// discriminant for fractional changes
	int	dlimit;		// discriminant limit
	int	yinc;		// discriminant change per y change
	char	inc;		// increment
};

static void incrementalize_y(struct PolyConvexEdge *edge,struct PolyVertex *top,struct PolyVertex *bottom);
static void increment(struct PolyConvexEdge *edge);

void vid13_polyConvex(BMP bm,struct PolyVertex *vert,int n,color c) {
	struct PolyVertex *vp;
	struct PolyConvexEdge l,r;
	int i,top;
	int y;
	int rem;	// # of vertices remaining
	int li,ri,ly,ry;

	////////////////////

	c = vid_ctran(c);

	// check for the easiest situations:
	//   <0 vertices: nonsense polygon
	//    0 vertices: no polygon
	//    1 vertex:   single dot
	//    2 vertices: line

	if (n<3) {
		if (n==2) vd->line(bm,vert[0].x,vert[0].y,vert[1].x,vert[1].y,c);
		else if (n==1) vd->plot(bm,vert[0].x,vert[0].y,c);
		return;
	}

	// try to find the highest (lowest) vertex

	vp	= vert;
	top	= 0;
	y	= vp++ -> y;
	for(i=1 ; i<n ; ++i,++vp) {
		if (vp->y < y) {
			y=vp->y;
			top = i;
		}
	}

	// initialize for scan conversion

	rem	= n;
	ly	= y-1;
	ry	= y-1;
	li	= top;
	ri	= top;

	while(rem>0) {
		while(ly<=y && rem>0) {
			--rem;
			i = li-1;
			if (i<0) i=n-1;
			incrementalize_y(&l,&vert[li],&vert[i]);
			ly = vert[i].y;
			li = i;
		}
		while(ry<=y && rem>0) {
			--rem;
			i = ri+1;
			if (i>=n) i=0;
			incrementalize_y(&r,&vert[ri],&vert[i]);
			ry = vert[i].y;
			ri=i;
		}
		while(y<ly && y<ry) {
			vd->hline(bm,l.x,r.x,y++,c);
			increment(&l);
			increment(&r);
		}
	}
}

static void	incrementalize_y(struct PolyConvexEdge *edge,struct PolyVertex *top,struct PolyVertex *bottom) {
	int dx,dy;

	////////////////

	edge->x		= top->x;

	if (top->y == bottom->y)
		return;		// one pixel high - what's the point computing?

	dx = bottom->x - top->x;
	dy = bottom->y - top->y;

	edge->idx	= dx/dy;
	edge->dlimit	= dy;
	edge->disc	= dy>>1;
	edge->yinc	= dx - edge->idx * dy;
	if (edge->yinc<0) edge->yinc = -edge->yinc;
	edge->inc	= (dx>0)-(dx<0);
}

static void	increment(struct PolyConvexEdge *edge) {
	edge->x		+= edge->idx;
	edge->disc 	-= edge->yinc;
	if (edge->disc<=0) {
		edge->disc += edge->dlimit;
		edge->x	   += edge->inc;
	}
}


