#define	VGAPAINT_VIDEO

#include <dos.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <math.h>

#include <a/types.h>
#include <a/mem.h>

#include "vgapaint.h"
#include "video.h"
#include "io.h"

#pragma off (check_stack)

static uchar palette_shadow[768];

void vid_setmode(int mode) {
	regs.h.al=mode;
	regs.h.ah=0x00;
	int386(0x10,&regs,&regs);
}

void vid_cursmove(long x,long y) {
	regs.h.ah=0x02;
	regs.h.bh=0;
	regs.h.dh=y;
	regs.h.dl=x;
	int386(0x10,&regs,&regs);
}

void vid_initBitMap(struct BitMap *bm,uchar *vm,long x,long y,long xs,long ys,long bpr,uchar *palette) {
	bm->vm=vm;
	bm->x1=x;
	bm->y1=y;
	bm->x2=x+xs;
	bm->y2=y+ys;
	bm->width=xs;
	bm->height=ys;
	bm->bpr=bpr;
	bm->zbpr=bpr<<2;
	bm->size=bpr*ys;
	bm->syncrect=FALSE;
	bm->xbias=bm->ybias=0;
	bm->pal = palette;
	bm->xdpi = 80;
	bm->ydpi = 72;
}

void vid_cloneBitMap(struct BitMap *bm,struct BitMap *old,long x,long y,long xo,long yo,long xs,long ys) {
	vid_initBitMap(bm,old->vm+x+y*old->bpr,xo,yo,xs,ys,old->bpr,old->pal);
	bm->xbias = old->xbias + x;
	bm->ybias = old->ybias + y;
	bm->syncrect = old->syncrect;
}

void vid13_set(struct BitMap *bm,color c) {
	c=vid_ctran(c);
	vid_tap_rect(bm,bm->x1,bm->y1,bm->width,bm->height);
	if (bm->bpr == bm->width)	/* if the bitmap occupies the full width */
		qset(bm->vm+bm->x1+bm->y1*bm->bpr,((long)c<<24)+((long)c<<16)+(c<<8)+c,bm->size);
	else				/* do it the slow way */
		vid13_bar(bm,bm->x1,bm->y1,bm->width,bm->height,c);
}

void vid13_plot(struct BitMap *bm,long x,long y,color c) {
	vid_tap_point(bm,x,y);
	if (x>=bm->x1)
		if (y>=bm->y1)
			if (x<bm->x2)
				if (y<bm->y2) {
					uchar *vm=bm->vm+bm->bpr*y+x;
					if (c<0)
						*vm^=(uchar)c;
					else
						*vm=c;
				};
}

color vid13_point(struct BitMap *bm,long x,long y) {
	if (x>=bm->x1)
		if (y>=bm->y1)
			if (x<bm->x2)
				if (y<bm->y2)
					return *(bm->vm+bm->bpr*y+x);
	return -1;
}

boolean vid_rectclip(struct BitMap *dst,long *x1,long *y1,long *xs,long *ys) {
	long dif;
	if (*x1<dst->x1) { dif=dst->x1-*x1; *x1+=dif; *xs-=dif; };
	if (*y1<dst->y1) { dif=dst->y1-*y1; *y1+=dif; *ys-=dif; };
	if (*x1+*xs>dst->x2) *xs=dst->x2-*x1;
	if (*y1+*ys>dst->y2) *ys=dst->y2-*y1;
	if (*xs>0)
		if (*ys>0)
			return TRUE;
	return FALSE;
}

boolean vid_rectclip2(struct BitMap *src,long *x1,long *y1,struct BitMap *dst,
	long *x2,long *y2,long *xs,long *ys) {
	long dif;
	if (*x1<src->x1) { dif=src->x1-*x1; *x1+=dif; *x2+=dif; *xs-=dif; };
	if (*y1<src->y1) { dif=src->y1-*y1; *y1+=dif; *y2+=dif; *ys-=dif; };
	if (*x2<dst->x1) { dif=dst->x1-*x2; *x1+=dif; *x2+=dif; *xs-=dif; };
	if (*y2<dst->y1) { dif=dst->y1-*y2; *y1+=dif; *y2+=dif; *ys-=dif; };
	if (*x1+*xs>src->x2) *xs=src->x2-*x1;
	if (*y1+*ys>src->y2) *ys=src->y2-*y1;
	if (*x2+*xs>dst->x2) *xs=dst->x2-*x2;
	if (*y2+*ys>dst->y2) *ys=dst->y2-*y2;
	if (*xs>0)
		if (*ys>0)
			return TRUE;
	return FALSE;
}

void vid13_bar(struct BitMap *dst,long x1,long y1,long xs,long ys,color c) {
	uchar *vm;
	long modulo;
	long xw4,xw1;
	ulong fillw;

	c=vid_ctran(c);

	if (!vid_rectclip(dst,&x1,&y1,&xs,&ys)) return;
	vid_tap_rect(dst,x1,y1,xs,ys);
	vm=dst->vm+y1*dst->bpr+x1;
	modulo=dst->bpr-xs;
	fillw=(c & 0xFF);
	fillw=(fillw<<8)+fillw;
	fillw=(fillw<<16)+fillw;
	xw4=xs>>2;
	xw1=xs & 3;

	if (c>=0)
		while(ys--) {
			vm=qset2(vm,fillw,xw4,xw1);
			vm+=modulo;
		}
	else
		while(ys--) {
			vm=qxor2(vm,fillw,xw4,xw1);
			vm+=modulo;
		};
}

void vid13_bitblt(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;
	long xw4,xw1;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(dst,x2,y2,xs,ys);
	xw4=xs>>2;
	xw1=xs & 3;
	modulo1=src->bpr;
	modulo2=dst->bpr;
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	if (src!=dst || vm1>vm2)
//		do {
//			qmove2(vm2,vm1,xw4,xw1);
//			vm1+=modulo1;
//			vm2+=modulo2;
//		} while(--ys);
		rectcopy(vm2,modulo2,vm1,modulo1,xs,ys);
	else {
		vm1+=xs+(ys-1)*src->bpr;
		vm2+=xs+(ys-1)*dst->bpr;
		do {
			qmove2r(vm2,vm1,xw4,xw1);
			vm1-=modulo1;
			vm2-=modulo2;
		} while(--ys);
				
	};
}

void vid13_bitbltxor(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;
	long xw4,xw1,t;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(dst,x2,y2,xs,ys);
	xw4=xs>>2;
	xw1=xs & 3;
	modulo1=src->bpr - xs;
	modulo2=dst->bpr - xs;
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;

	do {
		if (t=xw4) {
			do {
				*(ulong *)vm2 ^= *(ulong *)vm1;
				vm1 += sizeof(ulong);
				vm2 += sizeof(ulong);
			} while(--t);
		}
		if (t=xw1)
			do *vm2++ ^= *vm1++; while(--t);

		vm1+=modulo1;
		vm2+=modulo2;
	} while(--ys);
}

void vid13_bitbltcolor(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys,uchar colr1,uchar colr2) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(dst,x2,y2,xs,ys);
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	modulo1=src->bpr;
	modulo2=dst->bpr;
	while(ys--) {
		qmoveclrb(vm2,vm1,xs,colr1,colr2);
		vm1+=modulo1;
		vm2+=modulo2;
	};
}

void vid13_bitblttrans(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys,uchar c) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(dst,x2,y2,xs,ys);
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	modulo1=src->bpr;
	modulo2=dst->bpr;
	while(ys--) {
		qmovetrnb(vm2,vm1,xs,c);
		vm1+=modulo1;
		vm2+=modulo2;
	};
}

void vid13_bitbltswap(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;
	long xw4,xw4b,xw1;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(src,x1,y1,xs,ys);
	vid_tap_rect(dst,x2,y2,xs,ys);
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	xw4=xs>>2;
	xw1=xs & 3;
	modulo1=src->bpr-(xw4b=(xw4<<2));
	modulo2=dst->bpr-(xw4<<2);
	while(ys--) {
		if (xw4) {
			qswapl((ulong *)vm2,(ulong *)vm1,xw4);
			vm1+=xw4b;
			vm2+=xw4b;
		};
		if (xw1) qswapb(vm2,vm1,xw1);
		vm1+=modulo1;
		vm2+=modulo2;
	};
}

void vid13_bitbltswaptrans(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	long x2,long y2,long xs,long ys,uchar c) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(src,x1,y1,xs,ys);
	vid_tap_rect(dst,x2,y2,xs,ys);
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	modulo1=src->bpr;
	modulo2=dst->bpr;
	while(ys--) {
		qswaptrnb(vm2,vm1,xs,c);
		vm1+=modulo1;
		vm2+=modulo2;
	};
}

void vid13_bitbltswapmask(struct BitMap *src,long x1,long y1,struct BitMap *dst,
	uchar *mask,long x2,long y2,long xs,long ys,uchar c) {

	uchar *vm1,*vm2;
	long modulo1,modulo2;

	if (!vid_rectclip2(src,&x1,&y1,dst,&x2,&y2,&xs,&ys)) return;
	vid_tap_rect(src,x1,y1,xs,ys);
	vid_tap_rect(dst,x2,y2,xs,ys);
	vm1=src->vm+y1*src->bpr+x1;
	vm2=dst->vm+y2*dst->bpr+x2;
	mask+=y2*dst->bpr+x2;
	modulo1=src->bpr;
	modulo2=dst->bpr;
	while(ys--) {
		qswapmskb(vm2,vm1,xs,mask,c);
		vm1+=modulo1;
		vm2+=modulo2;
		mask+=modulo2;
	};
}

void vid13_bltpattern(struct BitMap *dst,ushort *data,long x1,long y1,long xs,long ys,color fore,color back) {
	uchar *vm;
	long xstemp;
	long vmod,dmod,dwid;
	long bitcnt;
	long preshift=0;
	ushort bitdata;

	if (xs<=0 || ys<=0 || x1+xs<dst->x1 || y1+ys<dst->y1)
		return;

	dwid=(xs+15)>>4;

	if (x1<dst->x1) {
		xstemp=dst->x1-x1;
		data+=(xstemp>>4);
		xs-=xstemp;
		preshift=xstemp & 15;
		x1=dst->x1;
	};
	if (x1+xs>=dst->x2)
		if ((xs=dst->x2-x1)<=0) return;

	if (y1<dst->y1) {
		xstemp=dst->y1-y1;
		data+=dwid*xstemp;
		ys-=xstemp;
		y1=dst->y1;
	};
	if (y1+ys>=dst->y2)
		if ((ys=dst->y2-y1)<=0) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	vm=dst->vm+x1+y1*dst->bpr;
	vmod=dst->bpr-xs;
	dmod=dwid-((preshift+xs+15)>>4);
	xstemp=xs;

	while(ys--) {
		bitcnt=16-preshift;
		bitdata=*data++;
		if (preshift) bitdata<<=preshift;
		for (xs=xstemp;xs;xs--) {
			if (!bitcnt--) {
				bitcnt=15;
				bitdata=*data++;
			};
			if ((signed short)bitdata < 0) *vm++=fore; else *vm++=back;
			bitdata<<=1;
		};
		vm+=vmod;
		data+=dmod;
	};
}

void vid13_bltpattern3d(struct BitMap *dst,ulong *data,long x1,long y1,long xs,long ys) {
	uchar *vm;
	long xstemp;
	long vmod,dmod,dwid;
	long bitcnt;
	long preshift=0;
	long preshift2;
	ulong bitdata;

	if (xs<=0 || ys<=0 || x1+xs<dst->x1 || y1+ys<dst->y1)
		return;

	dwid=(xs+15)>>4;

	if (x1<dst->x1) {
		xstemp=dst->x1-x1;
		data+=(xstemp>>4);
		xs-=xstemp;
		preshift=(xstemp & 15)<<1;
		x1=dst->x1;
	};
	if (x1+xs>=dst->x2)
		if ((xs=dst->x2-x1)<=0) return;

	if (y1<dst->y1) {
		xstemp=dst->y1-y1;
		data+=dwid*xstemp;
		ys-=xstemp;
		y1=dst->y1;
	};
	if (y1+ys>=dst->y2)
		if ((ys=dst->y2-y1)<=0) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	vm=dst->vm+x1+y1*dst->bpr;
	vmod=dst->bpr-xs;
	dmod=dwid-((preshift+xs+15)>>4);
	xstemp=xs;
	preshift2=preshift<<1;

	while(ys--) {
		bitcnt=16-preshift;
		bitdata=*data++;
		if (preshift) bitdata<<=preshift2;
		for (xs=xstemp;xs;xs--) {
			if (!bitcnt--) {
				bitcnt=15;
				bitdata=*data++;
			};
			bitdata=_lrotl(bitdata,2);
			*vm++=ctran(bitdata & 3);
		};
		vm+=vmod;
		data+=dmod;
	};
}

void vid13_bltpattern3dscale(struct BitMap *dst,ulong *data,long x1,long y1,long xs,long ys,long xmult,long ymult) {
	uchar *vm;
	ulong	*data_line;
	long xstemp;
	long vmod,dwid;
	long bitcnt;
	long preshift=0;
	long preshift2;
	ulong bitdata;

	slong	xaccum=0,yaccum=0,xaccumt,yaccumt;
	ulong	xadd,yadd;

	if (xmult<=0 || ymult<=0 || xs<=0 || ys<=0) return;

	dwid=(xs+15)>>4;

	xadd = (0x7FFFFFFFUL+xmult)/xmult;
	yadd = (0x7FFFFFFFUL+ymult)/ymult;
	xs *= xmult;
	ys *= ymult;

	if (x1+xs<dst->x1 || y1+ys<dst->y1)
		return;

	if (x1<dst->x1) {
		xstemp=dst->x1-x1;
		data+=((xstemp/xmult)>>4);
		xs-=xstemp;
		preshift=((xstemp/xmult) & 15)<<1;
		x1=dst->x1;
		xaccum = xadd*(xstemp%xmult);
	};
	if (x1+xs>=dst->x2)
		if ((xs=dst->x2-x1)<=0) return;

	if (y1<dst->y1) {
		xstemp=dst->y1-y1;
		data += dwid*(xstemp/ymult);
		ys-=xstemp;
		y1=dst->y1;
		yaccum = yadd*(xstemp%ymult);
	};
	if (y1+ys>=dst->y2)
		if ((ys=dst->y2-y1)<=0) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	vm=dst->vm+x1+y1*dst->bpr;
	vmod=dst->bpr-xs;
	xstemp=xs;
	preshift2=preshift<<1;

	data_line = data;

	while(ys) {
		yaccumt=yaccum;
		while(ys && yaccumt>=0) {
			--ys;
			yaccumt += yadd;

			data = data_line;
			bitcnt=16-preshift;
			bitdata=*data++;
			if (preshift) bitdata<<=preshift2;

			xs=xstemp;
			xaccumt=xaccum;
			while(xs) {
				if (!bitcnt--) {
					bitcnt=15;
					bitdata=*data++;
				};
				bitdata=_lrotl(bitdata,2);
				while(xs && xaccumt>=0) {
					--xs;
					xaccumt += xadd;
					*vm++=ctran(bitdata & 3);
//					vid_printf6(&vidbm,0,0,7,0,"%08lx %5d\n",xaccumt,xs);
//					getch();
				}
				xaccumt=0;
			};
			vm+=vmod;
		}
		yaccum=0;
		data_line += dwid;
	};
}

void vid13_text6lim(struct BitMap *dst,long x,long y,char *s,color fore,color back,long lim) {
	uchar *vm=(uchar *)dst->vm;
	uchar *vm2;
	uchar *fontptr;
	long rowc,colc,rowt=8;
	uchar c;
	uchar fontc;
	long x2,preshift=0;

	fore=vid_ctran(fore);
	back=vid_ctran(back);

	vm+=dst->bpr*y+x;
	if (y<dst->y1-8) return;
	else if (y>=dst->y2) return;
	else if (x>=dst->x2) return;
	else if (y<dst->y1) {
		preshift=dst->y1-y;
		vm+=(dst->bpr)*(dst->y1-y);
		rowt=8+(y-dst->y1);
	} else if (y>dst->y2-8)
		rowt=dst->y2-y;

	if (x<dst->x1) x2=dst->width;
	else x2=dst->x2-x;

	vid_tap_rect(dst,x,y,strlen(s)*6-1,8);

	while((c=*s++) && x2>=0 && lim--) {
		fontptr=&font6[6*c];
		for (colc=(*s)?6:5;colc && x2>0;colc--,x++,x2--) {
			vm2=vm++;
			fontc=*fontptr++;
			if (x>=dst->x1) {
				if (preshift) fontc<<=preshift;
				for (rowc=rowt;rowc;rowc--) {
					*vm2=((schar)fontc<0)?fore:back;
					fontc<<=1;
					vm2+=dst->bpr;
				};
			};
		};
	};
}

void vid13_text6trans(struct BitMap *dst,long x,long y,char *s,color cl) {
	uchar *vm=(uchar *)dst->vm;
	uchar *vm2;
	uchar *fontptr;
	long rowc,colc,rowt=8;
	uchar c;
	uchar fontc;
	long x2,preshift=0;

	cl=vid_ctran(cl);

	vm+=dst->bpr*y+x;
	if (y<dst->y1-8) return;
	else if (y>=dst->y2) return;
	else if (x>=dst->x2) return;
	else if (y<dst->y1) {
		preshift=dst->y1-y;
		vm+=(dst->bpr)*(dst->y1-y);
		rowt=8+(y-dst->y1);
	} else if (y>dst->y2-8)
		rowt=dst->y2-y;

	if (x<dst->x1) x2=dst->width;
	else x2=dst->x2-x;

	vid_tap_rect(dst,x,y,strlen(s)*6-1,8);

	while((c=*s++) && x2>=0) {
		fontptr=&font6[6*c];
		for (colc=(*s)?6:5;colc && x2>0;colc--,x++,x2--) {
			vm2=vm++;
			fontc=*fontptr++;
			if (x>=dst->x1) {
				if (preshift) fontc<<=preshift;
				for (rowc=rowt;rowc;rowc--) {
					if ((schar)fontc<0) *vm2=cl;
					fontc<<=1;
					vm2+=dst->bpr;
				};
			};
		};
	};
}


void vid13_hline(struct BitMap *dst,long x1,long x2,long y,color c) {
	long t;
	uchar *vm;
	if (x2<x1) { t=x1; x1=x2; x2=t; };
	if (y<dst->y1 || y>=dst->y2) return;
	if (x1>=dst->x2 || x2<dst->x1) return;
	if (x1<dst->x1) x1=dst->x1;
	if (x2>=dst->x2) x2=dst->x2-1;
	vid_tap_rect2(dst,x1,y,x2,y);
	vm=dst->vm+y*dst->bpr+x1;
	if(c<0) qxor(vm,(uchar)c,x2+1-x1);
	else	qset(vm,(uchar)c,x2+1-x1);
}

void vid13_vline(struct BitMap *dst,long x,long y1,long y2,color c) {
	long t;
	uchar *vm;
	if (y2<y1) { t=y1; y1=y2; y2=t; };
	if (x<dst->x1 || x>=dst->x2) return;
	if (y1>=dst->y2 || y2<dst->x1) return;
	if (y1<dst->y1) y1=dst->y1;
	if (y2>=dst->y2) y2=dst->y2-1;
	vid_tap_rect2(dst,x,y1,x,y2);
	vm=dst->vm+y1*dst->bpr+x;
	if(c<0) {
		while(y1++<=y2) {
			*vm^=c;
			vm+=dst->bpr;
		};
	} else	while(y1++<=y2) {
			*vm=c;
			vm+=dst->bpr;
		};
}

/* this is really nothing more than a quick and dirty hack for palette.c */
/* thus it does minimal clipping */

void vid13_quickfill(BMP dst,coord x1,coord y1,coord xs,coord ys) {
	uchar *vm=(uchar *)dst->vm;
	long xstemp,modulo;

	if (!vid_rectclip(dst,&x1,&y1,&xs,&ys)) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	vm+=dst->bpr*y1+x1;
	xs--;
	modulo=dst->bpr-xs;

	while(ys--) {
		xstemp=xs;
		while(xstemp--) {
			if (*vm++) break;
		};
		if (xstemp>=0) while(xstemp--)
			if (!*vm++)
				if (*vm)
					vm[-1]=vm[-2];
		vm+=modulo;
	};
}

void vid13_flipx(BMP dst,coord x1,coord y1,coord xs,coord ys) {
	register uchar *vm;
	long bpr;

	if (!vid_rectclip(dst,&x1,&y1,&xs,&ys)) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	bpr=dst->bpr;
	vm=dst->vm+x1+y1*bpr;

	do {
		asm_flipx(vm,xs);
		vm+=bpr;
	} while(--ys);
}

void vid13_flipy(BMP dst,coord x1,coord y1,coord xs,coord ys) {
	uchar *vm1,*vm2;
	long bpr;

	if (!vid_rectclip(dst,&x1,&y1,&xs,&ys)) return;

	vid_tap_rect(dst,x1,y1,xs,ys);
	bpr=dst->bpr;
	vm1=dst->vm+x1+y1*bpr;
	vm2=vm1+(ys-1)*bpr;

	if (!(ys>>=1)) return;

	do {
		qswap2((ulong *)vm1,(ulong *)vm2,xs);
		vm1+=bpr;
		vm2-=bpr;
	} while(--ys);
}

/* WARNING: RotateRight does NO clipping. */

void vid13_rotateright(BMP src,coord x1,coord y1,BMP dst,coord x2,coord y2,coord xs,coord ys) {
	long smod,dmod,xst;
	uchar *vms,*vmd,*vmd2;

	if (xs<=0) return;
	if (ys<=0) return;

	vid_tap_rect(dst,x2,y2,ys,xs);
	vms =src->vm+x1+y1*src->bpr;
	vmd2=dst->vm+x2+y2*dst->bpr+ys;
	smod=src->bpr-xs;
	dmod=dst->bpr;

	do {
		vmd=--vmd2;
		xst=xs;
		do {
			*vmd=*vms++;
			vmd+=dmod;
		} while(--xst);
		vms+=smod;
	} while(--ys);
}

/* vid13_scale does full clipping for destination only */

void vid13_scale(BMP src,coord x1,coord y1,coord xs1,coord ys1,BMP dst,coord x2,coord y2,coord xs2,coord ys2) {
	uchar *vms,*vms2,*vmd;
	uchar c;
	ulong smod,dmod;
	long xacc,yacc=0;
	long xinit=0,xsize;
	long t;
	long xc,yc;

	xsize=xs2;

	if (x2<dst->x1) {
		xinit=xs1*(dst->x1-x2);
		t=xinit/xs2;
		xinit-=xs2*t;
		x1+=t;
		xsize-=(dst->x1-x2);
		x2=dst->x1;
	};

	if (x2+xsize>dst->x2)
		xsize-=(x2+xsize-dst->x2);

	yc=ys2;

	if (y2<dst->y1) {
		yacc=ys1*(dst->y1-y2);
		t=yacc/ys2;
		yacc-=ys2*t;
		y1+=t;
		yc-=(dst->y1-y2);
		y2=dst->y1;
	};

	if (y2+ys2>dst->y2)
		yc-=(y2+ys2-dst->y2);

	if (xsize<=0 || yc<=0) return;

	vid_tap_rect(dst,x2,y2,xsize,yc);
	vms2=src->vm+x1+y1*src->bpr;
	vmd =dst->vm+x2+y2*dst->bpr;
	smod=src->bpr;
	dmod=dst->bpr-xsize;

	do {
		if (ys1>=ys2) {	/* contract */
			while(yacc>0) {
				yacc-=ys2;
				vms2+=smod;
			};
		};
		vms=vms2;
		xc=xsize;
		xacc=xinit;
		if (xs1<xs2)	/* scale up */
			do {
				if (xacc>=0) {
					c=*vms++;
					xacc-=xs2;
				};
				xacc+=xs1;
				*vmd++=c;
			} while(--xc);
		else if (xs1>xs2)	/* scale down */
			do {
				while(xacc>0) {
					xacc-=xs2;
					++vms;
				};
				*vmd++=*vms;
				xacc+=xs1;
			} while(--xc);
		else	do { *vmd++=*vms++; } while(--xc);
		yacc+=ys1;
		if (ys1<ys2) {	/* expand */
			while(yacc>=ys2) {
				yacc-=ys2;
				vms2+=smod;
			};
		};
		vmd+=dmod;
	} while(--yc);
}

// vid13_scale_turbo does no clipping whatsoever

// scale up:
//    copy <integer> number of pixels
//    copy <fractional> pixel if accumulator dictates
//    next pixel

// scale down:
//    copy pixel
//    add <integer> advance
//    add <fractional> advance if accumulator dictates

#if 0
void vid13_scale_turbo(BMP src,coord x1,coord y1,coord xs1,coord ys1,BMP dst,coord x2,coord y2,coord xs2,coord ys2) {
	uchar *vms,*vms2,*vmd;
	uchar c;
	ulong smod,dmod;
	long xacc,yacc=0;
	long t;
	long xc,yc;

	vid_tap_rect(dst,x2,y2,xs2,ys2);
	vms2=src->vm+x1+y1*src->bpr;
	vmd =dst->vm+x2+y2*dst->bpr;
	smod=src->bpr;
	dmod=dst->bpr-xsize;

	yc=ys2;
	do {
		if (ys1>=ys2) {	/* contract */
			while(yacc>0) {
				yacc-=ys2;
				vms2+=smod;
			};
		};
		vms=vms2;
		xc=xs2;
		xacc=0;
		if (xs1<xs2)	/* scale up */
			do {
				c=*vms++;
				for(t=xint
				if (xacc>=0) {
					c=*vms++;
					xacc-=xs2;
				};
				xacc+=xs1;
				*vmd++=c;
			} while(--xc);
		else if (xs1>xs2)	/* scale down */
			do {
				while(xacc>0) {
					xacc-=xs2;
					++vms;
				};
				*vmd++=*vms;
				xacc+=xs1;
			} while(--xc);
		else	do { *vmd++=*vms++; } while(--xc);
		yacc+=ys1;
		if (ys1<ys2) {	/* expand */
			while(yacc>=ys2) {
				yacc-=ys2;
				vms2+=smod;
			};
		};
		vmd+=dmod;
	} while(--yc);
}
#endif

/* vid13_rotate: rotation!  Uses 3 pass shear method rewritten for
** pointer/coordinate accessing instead of three shears (clipping). */

#define rotatept(x0,y0) { long xx=(x0),yy=(y0); xacc=xx+yy*tan2; yacc=yy+xacc*sine; xacc+=yacc*tan2; }

void vid13_rotate(BMP src,long x1,long y1,long xs,long ys,long xc,long yc,BMP dst,long xc2,long yc2,double angle) {
	long *xoff,*xoffmem=NULL;
	long *yoff,*yoffmem=NULL;
	long *xp,*yp;
	uchar *vm1,*vm2;

	long x,y,x2,y2,xcnt,ycnt;
	long sbpr,dbpr;
	double tan2,sine;

	if (angle>=90) return;

	if (!(xoffmem=(long *)malloc((ys*4+1)*sizeof(long)))) goto abrt;
	if (!(yoffmem=(long *)malloc((xs*4+1)*sizeof(long)))) goto abrt;

	xoff=xoffmem+ys*2;
	yoff=yoffmem+xs*2;

	tan2=tan(angle/2);
	sine=sin(angle);

	/* construct x-coordinate shearing table */

	xp=xoffmem;
	for (y=-ys*2;y<=ys*2;y++)
		*xp++=floor(-tan2*(y+.5)+.5);

	yp=yoffmem;
	for(x=-xs*2;x<=xs*2;x++)
		*yp++=floor(sine*(x+.5)+.5);

	vm1=src->vm+src->bpr*y1+x1;
	vm2=dst->vm+dst->bpr*yc2+xc2;
	sbpr=src->bpr-xs;
	dbpr=dst->bpr;

	x1-=xc;
	y=y1-yc;
	for(ycnt=ys;ycnt;ycnt--) {
		x=x1+xoff[y];
		for (xcnt=xs;xcnt;xcnt--) {
			x2=x;
			y2=y+yoff[x2];
			x2+=xoff[y2];
			vm2[x2+y2*dbpr]=*vm1++;
			x++;
		};
		vm1+=sbpr;
		y++;
	};

abrt:
	if (xoffmem) free(xoffmem);
	if (yoffmem) free(yoffmem);
}

/* vid_rotate_size: return the bitmap size, in horizontal and vertical sizes,
**		    needed to hold a rotated bitmap of a given size
*/

void vid_rotate_size(long xs,long ys,long xc,long yc,long *xrs,long *yrs,
	long *xrc,long *yrc,double angle) {

	long x1,y1,x2,y2,x3,y3,x4,y4;
	long bx1,by1,bx2,by2;

/* calculate the coordinates for each of the 4 corners */

	x1=-xc;		y1=-yc;
	x2=xs-xc;	y2=-yc;
	x3=xs-xc;	y3=ys-yc;
	x4=-xc;		y4=ys-yc;

	vid_rotate_coordinate(&x1,&y1,angle);
	vid_rotate_coordinate(&x2,&y2,angle);
	vid_rotate_coordinate(&x3,&y3,angle);
	vid_rotate_coordinate(&x4,&y4,angle);

/* find the bounding box */

	bx1=min(min(x1,x2),min(x3,x4));
	by1=min(min(y1,y2),min(y3,y4));
	bx2=max(max(x1,x2),max(x3,x4));
	by2=max(max(y1,y2),max(y3,y4));

	*xrs=bx2+1-bx1;
	*yrs=by2+1-by1;
	*xrc=-bx1;
	*yrc=-by1;
}

/* vid_rotate_coordinate: internal, used to rotate a point */

void vid_rotate_coordinate(long *px,long *py,double angle) {
	long x,y;
	x=*px;
	y=*py;
	x+=floor(-tan(angle/2)*(y+.5)+.5);
	y+=floor(sin(angle)*(x+.5)+.5);
	x+=floor(-tan(angle/2)*(y+.5)+.5);
	*px=x;
	*py=y;
}

void vid13_tile(BMP src,BMP dst,coord x,coord y,coord xs2,coord ys2,coord xo,coord yo) {
	long ssize,dsize;
	long xleft,yleft;
	long xsize,ysize;
	long t;
	uchar *vmd,*vms;

	/* first, check for destination out-of-bounds; if so, correct */
	
	if (x<dst->x1) {
		xo+=(dst->x1-x);
		xs2-=(dst->x1-x);
		x=dst->x1;
	};
	if (y<dst->y1) {
		yo+=(dst->y1-y);
		ys2-=(dst->y1-y);
		y=dst->y1;
	};
	if (x+xs2>dst->x2) xs2=dst->x2-x;
	if (y+ys2>dst->y2) ys2=dst->y2-y;

	if (xs2<=0 || ys2<=0) return;

	vid_tap_rect(dst,x,y,xs2,ys2);

	if (xo<0) xo += src->width *(1-xo/src->width );
	if (yo<0) yo += src->height*(1-yo/src->height);
	xo %= src->width;
	yo %= src->height;

//	xo = xo - src->width * (xo / src->width);
//	if (xo<0) xo+=src->width;
//	yo = yo - src->height* (yo / src->height);
//	if (yo<0) yo+=src->height;

	/* do top left quarter first - always required */

	ssize=src->bpr;
	dsize=dst->bpr;

	vms=src->vm;
	vmd=dst->vm+x+y*dsize;

	xsize=min(xs2,src->width-xo);
	ysize=min(ys2,src->height-yo);
	xleft=xs2-xsize;
	yleft=ys2-ysize;

	if (xsize && ysize)
		rectcopy(vmd,dsize,vms+xo+yo*ssize,ssize,xsize,ysize);

	/* do top right quarter next if there is room */

	if (xo && xleft)
		rectcopy(vmd+xsize,dsize,vms+yo*ssize,ssize,min(xo,xleft),ysize);

	/* do bottom left quarter if necessary */

	if (yo && yleft)
		rectcopy(vmd+ysize*dsize,dsize,vms+xo,ssize,xsize,min(yo,yleft));

	/* do bottom right quarter if necessary */

	if (xo && yo && xleft && yleft)
		rectcopy(vmd+xsize+ysize*dsize,dsize,vms,ssize,min(xo,xleft),min(yo,yleft));

	/* fill horizontal row */

	xsize=src->width;
	ysize=src->height;
	xleft=xs2-xsize;
	yleft=ys2-ysize;
	vms=vmd;
	vmd=vms+src->width;

	while(xleft>0) {
		rectcopy(vmd,dsize,vms,dsize,t=min(xleft,xsize),min(ys2,ysize));
		xleft-=t;
		xsize+=t;
		vmd+=t;
	};

	vmd=vms+src->height*dsize;

	while(yleft>0) {
		rectcopy(vmd,dsize,vms,dsize,xs2,t=min(yleft,ysize));
		yleft-=t;
		ysize+=t;
		vmd+=t*dsize;
	};
}

ulong	vid13_rassize(coord xs,coord ys) {
	return vid13_bpr(xs)*ys;
}

ulong	vid13_bpr(coord xs) {
	return (xs+3)&0xFFFFFFFC;
}

void vid13_translate_asm(uchar *,ulong,ulong,ulong,uchar *);
#pragma aux vid13_translate_asm=	\
	"push ebp",			\
	"mov ebp,eax",			\
	"xor eax,eax",			\
	"l:",				\
	"mov ecx,ebx",			\
	"l2:",				\
	"mov al,[esi]",			\
	"mov al,[edi+eax]",		\
	"mov [esi],al",			\
	"inc esi",			\
	"dec ecx",			\
	"jne l2",			\
	"add esi,ebp",			\
	"dec edx",			\
	"jne l",			\
	"pop ebp",			\
	parm [esi] [ebx] [edx] [eax] [edi]	\
	modify exact [eax ecx edx esi];

void	vid13_translate(BMP dst,coord x1,coord y1,coord xs,coord ys,uchar *trnbuf) {
	if (!vid_rectclip(dst,&x1,&y1,&xs,&ys)) return;
	vid_tap_rect(dst,x1,y1,xs,ys);
	vid13_translate_asm(dst->vm+x1+y1*dst->bpr,xs,ys,dst->bpr-xs,trnbuf);
}

///////////////////// 3-POINT POLYGONS ////////////////////

void vid13_htrapezoid(BMP dst,coord x1,coord x2,coord y1,coord x3,coord x4,coord y2,color c) {
	coord	t;
	boolean	leftclip = FALSE;
	slong	x1_accum=0x8000,x2_accum=0x8000;
	slong	x1_fdelta,x2_fdelta;
	coord	x1_idelta,x2_idelta;
	div_t	dr;
	uchar	*vm,*vm1;
	ulong	cl;

//	printf("htrapezoid (%d,%d,%d)-(%d,%d,%d)\n",x1,x2,y1,x3,x4,y2);

	// make sure x1<=x2 && x3<=x4

	if (x1>x2) { t=x1; x1=x2; x2=t; }
	if (x3>x4) { t=x3; x3=x4; x4=t; }

	// make sure y1<=y2

	if (y1>y2) {
		t=y1; y1=y2; y2=t;
		t=x1; x1=x3; x3=t;
		t=x2; x2=x4; x4=t;
	}

//	printf("htrapezoid (%d,%d,%d)-(%d,%d,%d)\n",x1,x2,y1,x3,x4,y2);

	// if x1 and x3 are off the left, clip to a line
	// if x2 and x4 are off the right, clip to a line
	// if the above two conditions are true, call vd->bar and do a blockfill

	if (x1< dst->x1 && x3< dst->x1) leftclip = TRUE;
	if (x2>=dst->x2 && x4>=dst->x2) {
		if (leftclip) {
			vd->bar(dst,dst->x1,y1,dst->width,y2+1-y1,c);
			return;
		}

		x2_fdelta = x2_idelta = 0;
		x2 = dst->width;
	} else {
		// compute fixed-point accum for right side

		dr = div(x4-x2,y2-y1);

		x2_idelta = dr.quot;
		x2_fdelta = (dr.rem*0x10000L)/(y2-y1);

	}

	if (leftclip) {
		x1_fdelta = x1_idelta = 0;
		x1 = dst->x1;
	} else {
		// compute fixed-point accum for left side

		dr = div(x3-x1,y2-y1);

		x1_idelta = dr.quot;
		x1_fdelta = (dr.rem*0x10000L)/(y2-y1);
	}

	c = vid_ctran(c);
	cl = ((ulong)c & 0xff) * 0x01010101L;

	// clip polygon to top and bottom

	if (y2>dst->y2) y2=dst->y2;
	if (y1<dst->y1) {
		int yc = dst->y1 - y1;

		x1 += x1_idelta*yc; x1 += (x1_accum += x1_fdelta*yc)>>16; x1&=0x0000ffffL;
		x2 += x2_idelta*yc; x2 += (x2_accum += x2_fdelta*yc)>>16; x2&=0x0000ffffL;

		y1 = dst->y1;
	}

	// step down polygon

	vm = dst->vm + dst->bpr*y1;

	while(y1<=y2) {
		coord	xs;

		if (x1<dst->x1) {
			x1 = dst->x1;
			x1_fdelta = x1_idelta = 0;
		}
		if (x2>=dst->x2) {
			x2 = dst->x2-1;
			x2_fdelta = x2_idelta = 0;
		}
		xs = x2+1-x1;

//		printf("(%d,%d)-(%d,%d)\n",x1,y1,x2,y1);

		qset2(vm+x1,cl,xs>>2,xs&3);
		vm += dst->bpr;

		// advance bresenhams

		x1 += x1_idelta; x1 += (x1_accum += x1_fdelta)>>16; x1_accum&=0x0000ffffL;
		x2 += x2_idelta; x2 += (x2_accum += x2_fdelta)>>16; x2_accum&=0x0000ffffL;

		++y1;
	}
}

//////////////////////// DRIVER SPECIFIC OPERATIONS /////////////////////////////

static boolean vd13_buffered=FALSE;

boolean	vid13_init(enum VD_InitCode vdic,long x,long y) {
	void *realvm;

	switch(vdic) {
	case VD_STARTUP:
		vds=&vd13specs;
		if (vd13_buffered) {
			vidbm.vm=NULL;
			if (!(realvm=malloc(64000))) return FALSE;
			vid_initBitMap(&vidbm,(uchar *)realvm,0,0,320,200,320,piccol);
		} else {
			vid_initBitMap(&vidbm,(uchar *)0xA0000,0,0,320,200,320,piccol);
		};
	case VD_REINIT:
	case VD_TO13:
		vid_setmode(0x13);
		vid_initBitMap(&realvidbm,(uchar *)0xA0000,0,0,320,200,320,piccol);
		vidbm.syncrect=vd13_buffered;
		vidbm.syncx1=vidbm.syncy1=0x7FFF;
		vidbm.syncx2=vidbm.syncy2=-0x8000;
		maxclr=256;
		return TRUE;
	case VD_FROM13:
		return TRUE;
	case VD_CLOSEDOWN:
		if (vd13_buffered)
			if (vidbm.vm) free(vidbm.vm);
		return TRUE;
	case VD_GOBUFFER:
		vd13_buffered=TRUE;
		return TRUE;
	case VD_GODIRECT:
		vd13_buffered=FALSE;
		return TRUE;
	case VD_RESIZE:
		if (x!=320 || y!=200) return FALSE;
		else return TRUE;
	case VD_SETLINECOMPARE:
		x = x*2-1;
		_disable();
		outp(0x3d4,24);	outp(0x3d5,x & 0xff);
		outp(0x3d4,7); y=inp(0x3d5); outp(0x3d5,(x&0x100)?y|0x10:y&0xef);
		outp(0x3d4,9); y=inp(0x3d5); outp(0x3d5,(x&0x200)?y|0x40:y&0xbf);
		_enable();
		return FALSE;
	case VD_GETLINECOMPARE:
		_disable();
		outp(0x3d4,24);	x=(ulong)inp(0x3d5) & 0xff;
		outp(0x3d4,7); if (inp(0x3d5)&0x10) x+=0x100;
		outp(0x3d4,9); if (inp(0x3d5)&0x40) x+=0x200;
		_enable();
		return (x+1)/2;
	case VD_SETSTART:
		y*=80;
		y+=x;
		_disable();
		outp(0x3d4,12); outp(0x3d5,y>>8);
		outp(0x3d4,13); outp(0x3d5,y);
		_enable();
		return FALSE;
	case VD_GETSTART:
		_disable();
		outp(0x3d4,12); y = ((ulong)outp(0x3d5))<<8;
		outp(0x3d4,13); y += (ulong)outp(0x3d5);
		_enable();
		return (((ulong)y/320)<<16) + ((ulong)y % 320);
	};
	return FALSE;
}

void	vid13_rbitblt(coord x1,coord y1,coord xs,coord ys,uchar *mem,long dmod,boolean no_offset) {
	ulong offs=x1+y1*320;

	if (!no_offset) mem+=offs;

	rectcopy((uchar *)0xA0000+offs,320,mem,dmod,xs,ys);
};

void	vid13_rhline(coord x,coord y,uchar *mem,coord xs) {
	qmove((uchar*)0xA0000+x+320*y,mem,xs);
}

void	vid13_rhlineread(coord x,coord y,uchar *mem,coord xs) {
	qmove(mem,(uchar*)0xA0000+x+320*y,xs);
}

/* WARNING: Pointer functions are called from INTERRUPT! */

#pragma off (check_stack)

void	vid13_drawptr(struct MousePointerInfo *mpi,coord x,coord y) {
	uchar *vm=(uchar *)0xA0000;
	uchar *dm1=mpi->image;
	uchar *dm2=mpi->backup;
	uchar c;
	ulong xh=mpi->s_width,yh=mpi->s_height,xht;
	ulong vmod,dmod;

	if (x>=0)	vm+=x;
	if (y>=0)	vm+=(y * 320);

	mpi->vdp=vm;

	if (y<0) {
		yh+=y;
		dm1-=y*mpi->s_width;
	} else if (y>(200-mpi->s_height)) yh=200-y;

	if (x<0) {
		xh+=x;
		dm1-=x;
	} else if (x>(320-mpi->s_width)) xh=320-x;

	vmod=320-xh;
	dmod=mpi->s_width-xh;

	mpi->width	= xh;
	mpi->height	= yh;

	while(yh--) {
		for(xht=xh;xht;xht--) {
			*dm2++=*vm;
			if (c=*dm1++) *vm++=(c-1)?mpi->high:mpi->low; else vm++;
		};
		vm  += vmod;
		dm1 += dmod;
	};
}

void	vid13_eraseptr(struct MousePointerInfo *mpi) {
	uchar *vm=mpi->vdp;
	uchar *dm=mpi->backup;
	ulong x=mpi->width;
	ulong y=mpi->height;

	while(y--) {
		dm=qmovebes(vm,dm,x);
		vm+=320;
	};

}


void	vid13_hideptr(struct MousePointerInfo *mpi,boolean new_state) {
}

void	vid13_changeptr(struct MousePointerInfo *mpi) {
}

void	vid13_redrawptr(struct MousePointerInfo *mpi) {
	uchar *vm=mpi->vdp;
	uchar *dm1=mpi->image;
	uchar *dm2=mpi->backup;
	uchar c;
	ulong xh=mpi->width,yh=mpi->height,xht;
	ulong vmod,dmod;

	vmod=320-xh;
	dmod=mpi->s_width-xh;

	while(yh--) {
		for(xht=xh;xht;xht--) {
			if (c=*dm1++) *vm=(c-1)?mpi->high:mpi->low; else *vm=*dm2;
			++vm,++dm2;
		};
		vm  += vmod;
		dm1 += dmod;
	};
}


//////////////////////////// GENERIC OPERATIONS /////////////////////////////////



void	vid_tap_point(BMP dst,coord x,coord y) {
	if (!dst->syncrect) return;
	if (dst->syncx1>x) dst->syncx1=x;
	if (dst->syncy1>y) dst->syncy1=y;
	if (dst->syncx2<x) dst->syncx2=x;
	if (dst->syncy2<y) dst->syncy2=y;
}

void	vid_tap_rect(BMP dst,coord x1,coord y1,coord xs,coord ys) {
	coord x2,y2;

	if (!dst->syncrect) return;
	x2=x1+xs-1;
	y2=y1+ys-1;
	if (dst->syncx1>x1) dst->syncx1=x1;
	if (dst->syncy1>y1) dst->syncy1=y1;
	if (dst->syncx2<x2) dst->syncx2=x2;
	if (dst->syncy2<y2) dst->syncy2=y2;
}

void	vid_tap_rect2(BMP dst,coord x1,coord y1,coord x2,coord y2) {
	if (!dst->syncrect) return;
	if (dst->syncx1>x1) dst->syncx1=x1;
	if (dst->syncy1>y1) dst->syncy1=y1;
	if (dst->syncx2<x2) dst->syncx2=x2;
	if (dst->syncy2<y2) dst->syncy2=y2;
}

void	vid_tap_bitmap(BMP dst,BMP src) {
	coord xb,yb;

	if (!dst->syncrect) return;

	xb = src->xbias - dst->xbias;
	yb = src->ybias - dst->ybias;

	if (dst->syncx1>src->syncx1+xb) dst->syncx1=src->syncx1+xb;
	if (dst->syncy1>src->syncy1+yb) dst->syncy1=src->syncy1+yb;
	if (dst->syncx2<src->syncx2+xb) dst->syncx2=src->syncx2+xb;
	if (dst->syncy2<src->syncy2+yb) dst->syncy2=src->syncy2+yb;
}

void	vid_tap_reset(BMP dst) {
	dst->syncx1=dst->syncy1=0x7FFF;
	dst->syncx2=dst->syncy2=-0x8000;
}

// vid_refresh() needs to execute quickly, so the clipping code has been
// specialized

void	vid_refresh(void) {
	coord x1,y1,xs,ys;
	coord dif=0;

	if ((state.requester_count || !zoomstate) && !vidbm.syncrect) return;

	if (!state.requester_count && zoomstate)
		ZoomRefreshHandler();

	if (vidbm.syncrect) {
		x1=min(vidbm.syncx1,winbm->syncx1+winbm->xbias);
		y1=min(vidbm.syncy1,winbm->syncy1+winbm->ybias);
		if ((xs=max(vidbm.syncx2,winbm->syncx2+winbm->xbias)+1-x1)<0) goto xit;
		if ((ys=max(vidbm.syncy2,winbm->syncy2+winbm->ybias)+1-y1)<0) goto xit;

		if (x1<vidbm.x1) { dif=vidbm.x1-x1; x1+=dif; xs-=dif; };
		if (y1<vidbm.y1) { dif=vidbm.y1-y1; y1+=dif; ys-=dif; };
		if (x1+xs>vidbm.x2) xs=vidbm.x2-x1;
		if (y1+ys>vidbm.y2) ys=vidbm.y2-y1;
		if (xs>0)
			if (ys>0) {
				mouse_hide();
#if 0
				vd->bar(&vidbm,x1,y1,xs,ys,-1);
				vd->rbitblt(x1,y1,xs,ys,vidbm.vm,vidbm.bpr,FALSE);
				{
					ulong x=read_tick();
					while(x==read_tick()) kbhit();
				}
				vd->bar(&vidbm,x1,y1,xs,ys,-1);
#endif
				vd->rbitblt(x1,y1,xs,ys,vidbm.vm,vidbm.bpr,FALSE);
				mouse_show();
			};
	};
xit:
	vidbm.syncx1=vidbm.syncy1=0x7FFF;
	vidbm.syncx2=vidbm.syncy2=-0x8000;
	winbm->syncx1=winbm->syncy1=0x7FFF;
	winbm->syncx2=winbm->syncy2=-0x8000;
}



/////////////////////////// GENERAL DRAWING ROUTINES //////////////////////////////



#define	longswap(x,y)	{ t=x; x=y; y=t; }

/* implement the Cohen-Sutherland algorithm to quickly clip the line */

char vid_line_bounds(BMP dst,coord x,coord y) {
	char bc=0;
	if (y>=dst->y2) bc++;
	if (y< dst->y1) bc+=2;
	if (x>=dst->x2) bc+=4;
	if (x< dst->x1) bc+=8;
	return bc;
}

boolean vid_line_clip(BMP dst,coord *x1,coord *y1,coord *x2,coord *y2) {
	char c1,c2;
	coord t;
	coord xmax,ymax;
	boolean	swapit = FALSE;

	xmax=dst->x2-1;
	ymax=dst->y2-1;

	for(;;) {
		c1=vid_line_bounds(dst,*x1,*y1);
		c2=vid_line_bounds(dst,*x2,*y2);

		if (c1&c2) return TRUE;		/* reject line */
		if (!c1 & !c2) {		// accept line
			if (swapit) {
				longswap(*x1,*x2);
				longswap(*y1,*y2);
			}
			return FALSE;
		}

		if (!c1) {
			longswap(*x1,*x2);
			longswap(*y1,*y2);
			longswap(c1,c2);
			swapit = !swapit;
		};
		if (c1 & 1) {
			*x1+=(*x2-*x1)*(float)(ymax-*y1)/(*y2-*y1);
			*y1=ymax;
		};
		if (c1 & 2) {
			*x1+=(*x2-*x1)*(float)(dst->y1-*y1)/(*y2-*y1);
			*y1=dst->y1;
		};
		if (c1 & 4) {
			*y1+=(*y2-*y1)*(float)(xmax-*x1)/(*x2-*x1);
			*x1=xmax;
		};
		if (c1 & 8) {
			*y1+=(*y2-*y1)*(float)(dst->x1-*x1)/(*x2-*x1);
			*x1=dst->x1;
		};
	};
}

void vid_line(struct BitMap *dst,long x1,long y1,long x2,long y2,color c,void (*plot)(BMP,coord,coord,color),boolean nofirstdot) {
	long xt,yt,disc,t,dx,dy,xinc,yinc;
	boolean steep;

	c = vid_ctran(c);

	if (vid_line_clip(dst,&x1,&y1,&x2,&y2)) return;

	dx=abs(x2-x1); dy=abs(y2-y1);

	if (steep=(dx<dy)) {
		t=dx;
		dx=dy;
		dy=t;
	};

	xt=dy<<1;
	yt=xt-(dx<<1);
	disc=(dy<<1)-dx;

	if (steep) {
		if (y1>=y2) {
			--y2;
			xinc = -1;
		} else {
			++y2;
			xinc = +1;
		};

		yinc=(x1==x2)?0:((x1>x2)?-1:1);

		do {
			if (!nofirstdot)
				plot(dst,x1,y1,c);
			nofirstdot = FALSE;
			if (disc>0) {
				x1+=yinc;
				disc+=yt;
			} else  disc+=xt;

			y1 +=xinc;
		} while(y1!=y2);
	} else {
		if (x1>=x2) {
			--x2;
			xinc = -1;
		} else {
			++x2;
			xinc = +1;
		}

		yinc=(y1==y2)?0:((y1>y2)?-1:1);

		do {
			if (!nofirstdot)
				plot(dst,x1,y1,c);
			nofirstdot = FALSE;
			if (disc>0) {
				y1+=yinc;
				disc+=yt;
			} else  disc+=xt;

			x1 += xinc;
		} while(x1!=x2);
	};
}

void vid13_line(struct BitMap *dst,long x1,long y1,long x2,long y2,color c) {
	long	xt,yt,disc,t,dx,dy,yinc;
	boolean steep;
	ulong	bpr;
	uchar	*vm;
	int	i;

	c = vid_ctran(c);

	if (vid_line_clip(dst,&x1,&y1,&x2,&y2)) return;
	vid_tap_point(dst,x1,y1);
	vid_tap_point(dst,x2,y2);

	dx=abs(x2-x1); dy=abs(y2-y1);

	if (steep=(dx<dy)) {
		t=dx;
		dx=dy;
		dy=t;
	};

	xt=dy<<1;
	yt=-(dx<<1);
	disc=(dy<<1)-dx;

	if (steep) {
		if (y1>y2) {
			longswap(x1,x2);
			longswap(y1,y2);
		};

		yinc=(x1==x2)?0:((x1>x2)?-1:1);

		vm = dst->vm + x1 + y1 * (bpr=dst->bpr);

		i=y2+1-y1;

		if (c<0)
			do {
				*vm ^= (uchar)c;
				vm += bpr;
				if (disc>0) {
					vm += yinc;
					disc+=yt;
				}
				disc+=xt;
			} while(--i);
		else
			do {
				*vm = c;
				vm += bpr;
				if (disc>0) {
					vm += yinc;
					disc+=yt;
				}
				disc+=xt;
			} while(--i);
	} else {
		if (x1>x2) {
			longswap(x1,x2);
			longswap(y1,y2);
		};

		vm = dst->vm + x1 + y1 * (bpr=dst->bpr);

		yinc=(y1==y2)?0:((y1>y2)?-bpr:bpr);

		i=x2+1-x1;

		if (c<0)
			do {
				*vm++ ^= c;
				if (disc>0) {
					vm += yinc;
					disc+=yt;
				}
				disc+=xt;
			} while(--i);
		else
			do {
				*vm++ = c;
				if (disc>0) {
					vm += yinc;
					disc+=yt;
				}
				disc+=xt;
			} while(--i);
	};
}

void vid_box(struct BitMap *dst,long x1,long y1,long x2,long y2,color c) {
	long t;

	if (x2<x1) longswap(x1,x2);
	if (y2<y1) longswap(y1,y2);

	c = vid_ctran(c);

	vd->hline(dst,x1,x2,y1,c);
	if (y1!=y2) vd->hline(dst,x1,x2,y2,c);
	if (y2-y1>1) {
		vd->vline(dst,x1,y1+1,y2-1,c);
		if (x1!=x2)
			vd->vline(dst,x2,y1+1,y2-1,c);
	};
}

void vid_circle(BMP dst,long x,long y,long iy,color c) {
	long ie;
	long ix=0;

	vid_tap_rect2(dst,x-iy,y-iy,x+iy,y+iy);

	c = vid_ctran(c);

	ie = 2-iy;

	while (ix<=iy) {
		vd->plot(dst,x+ix,y+iy,c);
		vd->plot(dst,x-ix,y+iy,c);
		vd->plot(dst,x+ix,y-iy,c);
		vd->plot(dst,x-ix,y-iy,c);
		vd->plot(dst,x+iy,y+ix,c);
		vd->plot(dst,x-iy,y+ix,c);
		vd->plot(dst,x+iy,y-ix,c);
		vd->plot(dst,x-iy,y-ix,c);
		if (ie>0) {
			ie -= 2*iy-2;
			iy--;
		};
		ie += 2*ix + 3;
		ix++;
	};
}

void vid_circle_func(BMP dst,long x,long y,long r,color c,void (*plot)(BMP,coord,coord,color)) {
	long ie;
	long ix;
	long iy;

	c = vid_ctran(c);

	ix	= 0;
	ie	= 0;
	iy	= r;
	while (iy>=0) {
		plot(dst,x+iy,y-ix,c);
		if (ie>0)	ie-=2*--iy+1;
		else		ie+=2*++ix+1;
	}
	ix	= 0;
	ie	= 0;
	iy	= r;
	while (iy>=0) {
		plot(dst,x-ix,y-iy,c);
		if (ie>0)	ie-=2*--iy+1;
		else		ie+=2*++ix+1;
	}
	ix	= 0;
	ie	= 0;
	iy	= r;
	while (iy>=0) {
		plot(dst,x-iy,y+ix,c);
		if (ie>0)	ie-=2*--iy+1;
		else		ie+=2*++ix+1;
	}
	ix	= 0;
	ie	= 0;
	iy	= r;
	while (iy>=0) {
		plot(dst,x+ix,y+iy,c);
		if (ie>0)	ie-=2*--iy+1;
		else		ie+=2*++ix+1;
	}
}

void vid_ellipse(BMP dst,long xc,long yc,long a,long b,long c) {
	long x,y;
	long a2,b2,step1,step2,incr1,incr2,test,d;

	x=0;
	y=b;

	a2 = a*a;
	b2 = b*b;
	step1 = a2*4;
	step2 = b2*4;
	test = a2*b;
	d=2*(b2-test)+a2;
	incr1=step2+2*b2;
	incr2=step1-test*4;

	vid13_plot(dst,xc+x,yc-y,c);
	vid13_plot(dst,xc+(x++),yc+y,c);

	while(test>0) {
		vid13_plot(dst,xc-x,yc-y,c);
		vid13_plot(dst,xc+x,yc-y,c);
		vid13_plot(dst,xc-x,yc+y,c);
		vid13_plot(dst,xc+(x++),yc+y,c);
		if (d>=0) {
			d+=incr2;
			incr2+=step1;
			y--;
			test-=a2;
		};
		test -= b2;
		d += incr1;
		incr1+=step2;
	};

	y=0;
	x=a;
	test = b2*a;
	d=2*(a2-test)+b2;
	incr1=step1+2*a2;
	incr2=step2-4*test;

	vid13_plot(dst,xc-x,yc+y,c);
	vid13_plot(dst,xc+x,yc+(y++),c);

	while(test>0) {
		vid13_plot(dst,xc-x,yc-y,c);
		vid13_plot(dst,xc+x,yc-y,c);
		vid13_plot(dst,xc-x,yc+y,c);
		vid13_plot(dst,xc+x,yc+(y++),c);
		if (d>=0) {
			d+=incr2;
			incr2 += step2;
			x--;
			test-=b2;
		};
		test-=a2;
		d+=incr1;
		incr1+=step1;
	};
}

void vid_ellipse_func(BMP dst,long xc,long yc,long a,long b,long c,void (*plot)(BMP,coord,coord,color)) {
	long x,y;
	long a2,b2,step1,step2,incr1,incr2,test,d;

	x=0;
	y=b;

	a2 = a*a;
	b2 = b*b;
	step1 = a2*4;
	step2 = b2*4;
	test = a2*b;
	d=2*(b2-test)+a2;
	incr1=step2+2*b2;
	incr2=step1-test*4;

	plot(dst,xc+x,yc-y,c);
	plot(dst,xc+(x++),yc+y,c);

	while(test>0) {
		plot(dst,xc-x,yc-y,c);
		plot(dst,xc+x,yc-y,c);
		plot(dst,xc-x,yc+y,c);
		plot(dst,xc+(x++),yc+y,c);
		if (d>=0) {
			d+=incr2;
			incr2+=step1;
			y--;
			test-=a2;
		};
		test -= b2;
		d += incr1;
		incr1+=step2;
	};

	y=0;
	x=a;
	test = b2*a;
	d=2*(a2-test)+b2;
	incr1=step1+2*a2;
	incr2=step2-4*test;

	plot(dst,xc-x,yc+y,c);
	plot(dst,xc+x,yc+(y++),c);

	while(test>0) {
		plot(dst,xc-x,yc-y,c);
		plot(dst,xc+x,yc-y,c);
		plot(dst,xc-x,yc+y,c);
		plot(dst,xc+x,yc+(y++),c);
		if (d>=0) {
			d+=incr2;
			incr2 += step2;
			x--;
			test-=b2;
		};
		test-=a2;
		d+=incr1;
		incr1+=step1;
	};
}

void vid_setRGB6(uchar col,uchar r,uchar g,uchar b) {

	palette_shadow[col*3+0] = r;
	palette_shadow[col*3+1] = g;
	palette_shadow[col*3+2] = b;

	if (opt.gamma_adjust) {
		r=monitor_gamma[r];
		g=monitor_gamma[g+256];
		b=monitor_gamma[b+512];
	} else {
		r>>=2;
		g>>=2;
		b>>=2;
	};
	if (opt.nodirect_pal) {
		regs.w.ax = 0x1010;
		regs.w.bx = col;
		regs.h.dh = r;
		regs.h.ch = g;
		regs.h.cl = b;
		int386(0x10,&regs,&regs);
	} else {
		disable();
		outportb(0x3c8,col);
		outportb(0x3c9,r);
		outportb(0x3c9,g);
		outportb(0x3c9,b);
		enable();
	}
	if (ccycle_instd) CCycle_Reset();
}

void vid_getRGB6(uchar col,uchar *r,uchar *g,uchar *b) {
#if 0
	if (opt.nodirect_pal) {
		union REGS regs;

		regs.w.ax = 0x1015;
		regs.w.bx = col;
		int386(0x10,&regs,&regs);
		if (opt.gamma_adjust) {
			*r = bitmap_gamma[regs.h.dh];
			*g = bitmap_gamma[regs.h.ch];
			*b = bitmap_gamma[regs.h.cl];
		} else {
			*r = (regs.h.dh * 85L)/21;
			*g = (regs.h.ch * 85L)/21;
			*b = (regs.h.cl * 85L)/21;
		}
	} else {
		disable();
		outportb(0x3c7,col);
		if (opt.gamma_adjust) {
			*r=bitmap_gamma[inportb(0x3c9)];
			*g=bitmap_gamma[inportb(0x3c9)+64];
			*b=bitmap_gamma[inportb(0x3c9)+128];
		} else {
			*r = (inportb(0x3c9)*85L)/21;
			*g = (inportb(0x3c9)*85L)/21;
			*b = (inportb(0x3c9)*85L)/21;
		};
		enable();
	}
#else
	*r = palette_shadow[col*3+0];
	*g = palette_shadow[col*3+1];
	*b = palette_shadow[col*3+2];
#endif
}

void vid_loadRGB6(uchar col,uchar *mem,long cnt) {
	uchar *trnptr = paltrnmem;
	int	cnt2 = cnt;

	if (((ulong)col + cnt)>256) {
		memcpy(palette_shadow+3*col,mem,768-cnt*3);
		memcpy(palette_shadow,mem+(768-col*3),cnt-(256-col));
	} else
		memcpy(palette_shadow+3*col,mem,cnt*3);

	if (opt.gamma_adjust) {
		while(cnt--) {
			*trnptr++ = monitor_gamma[*mem++];
			*trnptr++ = monitor_gamma[*mem+++256];
			*trnptr++ = monitor_gamma[*mem+++512];
		};
	} else {
		cnt*=3;
		do {
			*trnptr++ = *mem++>>2;
		} while(--cnt);
	};
	if (opt.nodirect_pal) {
		struct RealInt rmi;

		rmi.w.ax = 0x1012;
		rmi.w.bx = col;
		rmi.w.cx = cnt2;
		rmi.w.dx = (ulong)paltrnmem & 0x0F;
		rmi.es = ((ulong)paltrnmem>>4);
		DPMIRealInt(0x10,&rmi);
	} else {
		disable();
		outportb(0x3c8,col);
		outportbs(0x3c9,paltrnmem,cnt2*3);
		enable();
	}
	if (ccycle_instd) CCycle_Reset();
}

void vid_saveRGB6(uchar col,uchar *mem,long cnt) {
#if 1
	uchar *trnptr = paltrnmem;	
	if (((ulong)col + cnt)>256) {
		memcpy(mem,palette_shadow+3*col,768-cnt*3);
		memcpy(mem+(768-col*3),palette_shadow,cnt-(256-col));
	} else
		memcpy(mem,palette_shadow+3*col,cnt*3);
#else
	if (opt.nodirect_pal) {
		struct RealInt rmi;

		rmi.w.ax = 0x1017;
		rmi.w.bx = col;
		rmi.w.cx = cnt;
		rmi.w.dx = (ulong)paltrnmem & 0x0F;
		rmi.es = ((ulong)paltrnmem>>4);
		DPMIRealInt(0x10,&rmi);
	} else {
		disable();
		outportb(0x3c7,col);
		inportbs(0x3c9,trnptr,cnt*3);
		enable();
	};
	if (opt.gamma_adjust) {
		while(cnt--) {
			*mem++=bitmap_gamma[*trnptr++      ];
			*mem++=bitmap_gamma[*trnptr++ +  64];
			*mem++=bitmap_gamma[*trnptr++ + 128];
		};
	} else {
		cnt*=3;
		do {
			*mem++=(*trnptr++ *85L)/21;
		} while(--cnt);
	};
#endif
}

void vid_setpreg(uchar pr,uchar col) {
	regs.w.ax=0x1000;
	regs.h.bh=col;
	regs.h.bl=pr;
	int386(0x10,&regs,&regs);
}

BLOCK vid_blocksave(BMP bm,long x1,long y1,long xs,long ys) {
	uchar *vm;
	BLOCK b;

	if (!(vm=(uchar *)malloc(xs*ys))) return NULL;
	if (!(b=(BLOCK)malloc(sizeof(struct Block)))) {
		free(vm);
		return NULL;
	};
	vid_initBitMap(&b->bm,vm,0,0,xs,ys,xs,bm->pal);
	vd->bitblt(&vidbm,x1,y1,&b->bm,0,0,xs,ys);
	b->x=x1;
	b->y=y1;
	b->xs=xs;
	b->ys=ys;
	return b;
}

void vid_blockrestore(BMP bm,BLOCK bl) {
	if (!bl) return;

	vd->bitblt(&bl->bm,0,0,&vidbm,bl->x,bl->y,bl->bm.width,bl->bm.height);

	free(&bl->bm.vm);
	free(bl);
}

void vid_printf6(BMP bm,long x,long y,color fore,color back,char *format, ...) {
	char str[128];
	va_list v;

	va_start(v,format);
	vsprintf(str,format,v);
	vd->text6(bm,x,y,str,fore,back);
	va_end(v);
}

void vid_text6(BMP bm,long x,long y,char *str,color fore,color back) {
	vd->text6lim(bm,x,y,str,fore,back,0x7FFF);
}

void vid_hblit(BMP bm,long x1,long y1,long xs,long ys,long dx) {
	if (dx<0)
		vd->bitblt(bm,x1-dx,y1,bm,x1,y1,xs+dx,ys);
	else if (dx>0)
		vd->bitblt(bm,x1,y1,bm,x1+dx,y1,xs-dx,ys);
}

void vid_vblit(BMP bm,long x1,long y1,long xs,long ys,long dy) {
	if (dy<0)
		vd->bitblt(bm,x1,y1-dy,bm,x1,y1,xs,ys+dy);
	else if (dy>0)
		vd->bitblt(bm,x1,y1,bm,x1,y1+dy,xs,ys-dy);
}

void vid_flash(long time) {
	uchar red,grn,blu;
	ulong	hit_tick;

	vid_getRGB6(0,&red,&grn,&blu);
	vid_setRGB6(0,255-red,255-grn,255-blu);
//	delay(time);

	hit_tick = read_tick()+time;

	while(read_tick()<hit_tick) kbhit();

	vid_setRGB6(0,red,grn,blu);
}

void vid_draw3d(BMP bm,long x1,long y1,long x2,long y2) {
	if (x1>x2) swap(long,x1,x2);
	if (y1>y2) swap(long,y1,y2);

	vd->hline(bm,x1,x2-1,y1,ST3);
	vd->vline(bm,x1,y1+1,y2,ST3);
	vd->hline(bm,x1+1,x2,y2,ST1);
	vd->vline(bm,x2,y1,y2-1,ST1);
	vd->plot(bm,x1,y2,ST2);
	vd->plot(bm,x2,y1,ST2);
}

void vid_inv3d(BMP bm,long x1,long y1,long x2,long y2) {
	if (x1>x2) swap(long,x1,x2);
	if (y1>y2) swap(long,y1,y2);

	vd->hline(bm,x1,x2-1,y1,ST1);
	vd->vline(bm,x1,y1+1,y2,ST1);
	vd->hline(bm,x1+1,x2,y2,ST3);
	vd->vline(bm,x2,y1,y2-1,ST3);
	vd->plot(bm,x1,y2,ST2);
	vd->plot(bm,x2,y1,ST2);
}

void vid_polyline(BMP bm,color c,long lcnt, ...) {
	va_list val;
	coord fx,fy,ox,oy,x,y;
	boolean close=FALSE;

	va_start(val,lcnt);
	fx=ox=va_arg(val,coord);
	fy=oy=va_arg(val,coord);
	if (lcnt & 0x80000000) {
		lcnt&=0x7FFFFFFF;
		close=TRUE;
	};
	while(lcnt--) {
		x=va_arg(val,coord);
		y=va_arg(val,coord);
		vd->line(bm,ox,oy,x,y,c);
		ox=x;	oy=y;
	};
	va_end(val);
	vd->line(bm,x,y,fx,fy,c);
}

/* vid_ctran: color translator */

color vid_ctran(color cp) {
	if ((cp & 0xFFFFFF00) == 0x80000100) return (((color *)&col)[cp & 0xFF]);
	else return cp;
};

/* vid_make_remap: obtain a remapping from one palette to another */

void vid_make_remap(uchar *srccol,uchar *destcol,uchar *mapmem,long src_colors,long dst_colors) {
#if 0
	uchar *dcl;
	long x,y;
	long rd,gd,bd,rdif,mdif,index;
	long r,g,b;

	/* do...while generates better code in Watcom C/386 */

	for(y=src_colors;y;y--) {
		r=*srccol++;
		g=*srccol++;
		b=*srccol++;
		dcl=destcol;
		mdif=0x7FFF;
		x=dst_colors;
		do {
			rd=r-*dcl++;
			gd=g-*dcl++;
			bd=b-*dcl++;
			rdif=rd*rd+gd*gd+bd*bd;
			if (rdif<mdif) {
				mdif=rdif;
				index=x;
			};
		} while(--x);
		*mapmem++=256-index;
	};
#endif

	struct FastPick *fp;
	int i;

	if (!(fp=FPick_Allocate(dst_colors))) return;

	FPick_Initialize(fp,0,dst_colors,destcol);

	for(i=src_colors;i;i--) {
		*mapmem++=FPick_Choose(fp,dst_colors,srccol[0],srccol[1],srccol[2]);
		srccol+=3;
	};

	FPick_Free(fp);
}

void Message(char *f, ...) {
	va_list val;
	char *tempbuf;
	long xsize=0,ysize=0;
	coord xs,ys;
	BLOCK save;

	if (!(tempbuf = malloc(2048))) return;

	va_start(val,f);
	vsprintf(tempbuf,f,val);
	va_end(val);

	// determine width and height of box

	{
		char *last	= tempbuf;
		char *s		= tempbuf;
		char c;
		int l;

		do {
			c = *s++;
			if (c=='\n' || !c) {
				l = s-last-1;
				if (l>xsize) xsize=l;
				++ysize;
				last = s;
			}
		} while(c);
	}

	// render the message box

	xs	= xsize*sysfontwidth + 4*XPOINT;
	ys	= ysize*sysfontheight + 4*YPOINT;

	mouse_hide();
	save = vd->blocksave(&vidbm,(vidbm.width-xs)>>1,(vidbm.height-ys)>>1,xs,ys);
	vd->bar(&vidbm,save->x,save->y,save->xs,save->ys,St2);
	vid_draw3d(&vidbm,save->x,save->y,save->x+save->xs-1,save->y+save->ys-1);

	{
		char *last	= tempbuf;
		char *s		= tempbuf;
		char c;
		int l;
		coord x		= save->x + 2*XPOINT;
		coord y		= save->y + 2*YPOINT;

		do {
			c = *s++;
			if (c=='\n' || !c) {
				l = s-last-1;

				SysTextLim(&vidbm,x+((xsize-l)*sysfontheight)/2,y,last,St0,St2,l);
				y += sysfontheight;

				last = s;
			}
		} while(c);
	}
	vd->refresh();
	mouse_show();

	while(!mouse_buttons() && !kbhit());
	mouse_wait_unkey();

	kbd_check();

	mouse_hide();
	vd->blockrestore(&vidbm,save);
	mouse_show();

	free(tempbuf);
}
