#ifndef	VGAPAINT_VIDVESA_H
#define	VGAPAINT_VIDVESA_H

#include "vgapaint.h"
#include "video.h"

#ifdef	VGAPAINT_VIDVESA_C
#define	EXTERN
#define	INIT(x)	=x
#else
#define	EXTERN	extern
#define	INIT(x)
#endif

#define	VIB_DAC8BIT		(1)
#define VIB_WEIRDVGA		(2)
#define VIB_DACNOISE		(4)

struct VbeInfoBlock {
	uchar	VbeSignature[4];
	ushort	VbeVersion;
	ulong	OemStringPtr;
	ulong	Capabilities;
	ulong	VideoModePtr;
	ushort	TotalMemory;
	ushort	OemSoftwareRev;
	ulong	OemVendorNamePtr;
	ulong	OemProductNamePtr;
	ulong	OemProductRevPtr;
	uchar	Reserved[222];
	uchar	Oemdata[256];
};

#define	VMI_MODE_SUPPORTED	1
#define	VMI_MODE_INFOVALID	2
#define	VMI_MODE_BIOSTEXT	4
#define	VMI_MODE_COLOR		8
#define	VMI_MODE_GRAPHICS	16
#define	VMI_MODE_NO_WINDOW	64
#define	VMI_MODE_LINEAR_ABLE	128

#define	VMI_WIN_SUPPORTED	1
#define	VMI_WIN_READABLE	2
#define	VMI_WIN_WRITABLE	4

#define	VMI_MEM_TEXT		0
#define	VMI_MEM_CGA		1
#define	VMI_MEM_HERCULES	2
#define	VMI_MEM_4PLANE		3
#define	VMI_MEM_PACKPIXEL	4
#define	VMI_MEM_MODEX		5

struct VESAModeInfo {
	ushort	mode_attr;		// mode attributes
	uchar	win_a_attr;		// A window attributes
	uchar	win_b_attr;		// B window attributes
	ushort	win_granularity;	// Window granularity
	ushort	win_size;		// Window size in KB
	ushort	win_a_segment;		// A window segment
	ushort	win_b_segment;		// B window segment
	ulong	win_page_routine;	// Paging routine, 16-bit far ptr
	ushort	bpr;			// Bytes per row

	/* optional */

	ushort	hres,vres;		// horiz and vert resolution
	uchar	cwid,chei;		// character size in pixels
	uchar	planes;			// bitplanes
	uchar	bpp;			// bits per pixel
	uchar	banks;			// # of banks
	uchar	mem_model;		// memory model
	uchar	bank_size;		// bank size
	uchar	page_count;		// number of pages
	uchar	reserved;

	uchar	red_mask_size;
	uchar	red_field_position;
	uchar	green_mask_size;
	uchar	green_field_position;
	uchar	blue_mask_size;
	uchar	blue_field_position;
	uchar	rsrv_mask_size;
	uchar	rsrv_field_position;
	uchar	directcolor_mode_info;

	ulong	phys_baseptr;		// PhysBasePtr NOT usable since it is a physical pointer
	ushort	offscreen_offset[2];
	ushort	offscreen_size;

	uchar	pad0[206];
};

enum VESAModeCheck {
	VMCHK_MODE_FINE		= 0,	// looks okay
	VMCHK_NOT_SUPPORTED	= 1,	// mode doesn't exist!!!
	VMCHK_NOT_GRAPHICS	= 2,	// mode isn't a graphics mode
	VMCHK_NO_READ_WINDOW	= 3,	// no read window (???)
	VMCHK_NO_WRITE_WINDOW	= 4,	// no write window (???)
	VMCHK_TINY_WINDOW	= 5,	// window is very, very small...
	VMCHK_WEIRD_WINDOW_SIZE	= 6,	// some number like 150 for window size
	VMCHK_TOO_SMALL		= 7,	// no low-low-low-res modes
	VMCHK_WRONG_DEPTH	= 8,	// depth should be 8
};

#define	VESA_640x400	0x100
#define	VESA_640x480	0x101
#define	VESA_800x600	0x103
#define	VESA_1024x768	0x105
#define	VESA_1280x1024	0x107

//enum VESAErr {
//	
//};

EXTERN struct VDevSpecs vdvesa_specs
#ifdef VGAPAINT_VIDVESA_C
={	640,400,640,400,640,400,640,400		}
#endif
;

void	VESA_Get_Info(ushort mode,struct VESAModeInfo *vmi);
boolean	VESA_Scan_Modes(void);
int	VESA_Check_Mode_Acceptable(struct VESAModeInfo *vmi,boolean set_mode);
boolean	vidvesa_init(enum VD_InitCode vdic,long x,long y);
void	vidvesa_rhline(coord x,coord y,uchar *dm1,coord xs);
void	vidvesa_rhlineread(coord x,coord y,uchar *dm1,coord xs);
void	vidvesa_rbitblt(coord x,coord y,coord xs,coord ys,uchar *dm1,ulong dmod,boolean no_offset);
void	vidvesa_rbitbltread(uchar *dm1,coord x,coord y,coord xs,coord ys,ulong dmod,boolean no_offset);
void	vidvesa_drawptr(struct MousePointerInfo *mpi,coord x,coord y);
void	vidvesa_eraseptr(struct MousePointerInfo *mpi);


EXTERN struct VDev vdvesa
#ifdef	VGAPAINT_VIDVESA_C
	={
		&vid13_set,
		&vid13_plot,
		&vid13_point,
		&vid13_bar,
		&vid13_bitblt,
		&vid13_bitbltcolor,
		&vid13_bitblttrans,
		&vid13_bitbltswap,
		&vid13_bitbltswaptrans,
		&vid13_bitbltswapmask,
		&vid13_bltpattern,
		&vid13_bltpattern3d,
		&vid_text6,		/* rely on limit text routine */
		&vid13_text6trans,
		&vid13_text6lim,
		&vid13_hline,
		&vid13_vline,
		&vid13_line,
		&vid_box,
		&vid_circle,
		&vid_ellipse,
		&vid_blocksave,
		&vid_blockrestore,
		&vid_hblit,
		&vid_vblit,
		&vid13_tile,
		&vid13_rassize,
		&vid13_bpr,
		&vidvesa_init,
		&vid_refresh,
		&vidvesa_rhline,
		&vid13_quickfill,
		&vidvesa_rbitblt,
		&vidvesa_drawptr,
		&vidvesa_eraseptr,
		&vidvesa_rhlineread
	}
#endif
;

EXTERN struct VESAModeInfo	vesa_minfo;

#undef	INIT
#undef	EXTERN

#endif

